import importlib
import re
import sys
from io import StringIO
from unittest.mock import MagicMock, patch

import pytest

import llm_accounting.backends.sqlite as sqlite_backend_module
from llm_accounting import backends
from llm_accounting.cli.main import main as cli_main


@patch("llm_accounting.cli.utils.get_accounting")
def test_select_aggregation(mock_get_accounting, test_db, capsys):
    """Test query with GROUP BY and aggregation"""
    mock_accounting_instance = MagicMock()
    mock_backend_instance = MagicMock()
    mock_accounting_instance.backend = mock_backend_instance
    mock_get_accounting.return_value = mock_accounting_instance
    mock_accounting_instance.__enter__.return_value = mock_accounting_instance
    mock_accounting_instance.__exit__.return_value = None
    mock_backend_instance.execute_query.return_value = [
        {'model': 'gpt-4', 'count': 2, 'total_input': 250},
        {'model': 'gpt-3.5', 'count': 2, 'total_input': 125}
    ]

    with patch.object(sys, 'argv', ['cli_main', "select", "--query", "SELECT model, COUNT(*) as count, SUM(prompt_tokens) as total_input FROM accounting_entries GROUP BY model", "--format", "table"]):
        cli_main()

    captured = capsys.readouterr()
    assert "gpt-4" in captured.out
    assert "gpt-3.5" in captured.out
    assert "2" in captured.out  # Verify group count

    assert re.search(r'gpt-4\W+2\W+250', captured.out, re.IGNORECASE)
    assert re.search(r'gpt-3\.5\W+2\W+125', captured.out, re.IGNORECASE)
    mock_accounting_instance.__exit__.assert_called_once()



import sqlite3
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import List

import pytest

from llm_accounting.backends.base import UsageEntry
from llm_accounting.backends.sqlite import SQLiteBackend
from llm_accounting.models.limits import UsageLimitDTO, LimitScope, LimitType, TimeInterval

# --- Tests for Project-based Limits ---

def test_insert_project_scope_limit(sqlite_backend: SQLiteBackend):
    """Test inserting a usage limit with PROJECT scope."""
    project_name = "Solaris"
    limit = UsageLimitDTO(
        scope=LimitScope.PROJECT.value,
        limit_type=LimitType.COST.value,
        max_value=100.0,
        interval_unit=TimeInterval.MONTH.value,
        interval_value=1,
        project_name=project_name
    )
    sqlite_backend.insert_usage_limit(limit)
    
    retrieved_limits = sqlite_backend.get_usage_limits(scope=LimitScope.PROJECT, project_name=project_name)
    assert len(retrieved_limits) == 1
    retrieved_limit = retrieved_limits[0]
    assert retrieved_limit.scope == LimitScope.PROJECT.value
    assert retrieved_limit.project_name == project_name
    assert retrieved_limit.max_value == 100.0

def test_get_usage_limits_filter_by_project_scope(sqlite_backend: SQLiteBackend):
    """Test filtering limits by PROJECT scope."""
    sqlite_backend.insert_usage_limit(UsageLimitDTO(scope=LimitScope.GLOBAL.value, limit_type=LimitType.COST.value, max_value=1000, interval_unit=TimeInterval.MONTH.value, interval_value=1))
    sqlite_backend.insert_usage_limit(UsageLimitDTO(scope=LimitScope.PROJECT.value, limit_type=LimitType.REQUESTS.value, max_value=500, interval_unit=TimeInterval.DAY.value, interval_value=1, project_name="ProjectX"))
    sqlite_backend.insert_usage_limit(UsageLimitDTO(scope=LimitScope.PROJECT.value, limit_type=LimitType.COST.value, max_value=200, interval_unit=TimeInterval.WEEK.value, interval_value=1, project_name="ProjectY"))
    
    project_limits = sqlite_backend.get_usage_limits(scope=LimitScope.PROJECT)
    assert len(project_limits) == 2
    assert all(limit.scope == LimitScope.PROJECT.value for limit in project_limits)
    
    project_x_limits = sqlite_backend.get_usage_limits(scope=LimitScope.PROJECT, project_name="ProjectX")
    assert len(project_x_limits) == 1
    assert project_x_limits[0].project_name == "ProjectX"
    assert project_x_limits[0].limit_type == LimitType.REQUESTS.value

def test_get_accounting_entries_for_quota_with_project_filter(sqlite_backend: SQLiteBackend):
    """Test get_accounting_entries_for_quota filtering by project_name."""
    now = datetime.now(timezone.utc)
    project_one = "ProjectOne"
    project_two = "ProjectTwo"

    # Entries for ProjectOne
    sqlite_backend.insert_usage(UsageEntry(model="gpt-4", cost=1.0, execution_time=1, project=project_one, timestamp=now - timedelta(minutes=10)))
    sqlite_backend.insert_usage(UsageEntry(model="gpt-4", cost=1.5, execution_time=1, project=project_one, timestamp=now - timedelta(minutes=5)))
    
    # Entries for ProjectTwo
    sqlite_backend.insert_usage(UsageEntry(model="gpt-4", cost=2.0, execution_time=1, project=project_two, timestamp=now - timedelta(minutes=10)))
    
    # Entry with no project
    sqlite_backend.insert_usage(UsageEntry(model="gpt-4", cost=0.5, execution_time=1, timestamp=now - timedelta(minutes=5))) # No project

    start_time = now - timedelta(hours=1)
    
    # Quota for ProjectOne
    cost_project_one = sqlite_backend.get_accounting_entries_for_quota(
        start_time=start_time, limit_type=LimitType.COST, project_name=project_one
    )
    assert cost_project_one == 2.5 # 1.0 + 1.5

    # Quota for ProjectTwo
    cost_project_two = sqlite_backend.get_accounting_entries_for_quota(
        start_time=start_time, limit_type=LimitType.COST, project_name=project_two
    )
    assert cost_project_two == 2.0

    # Quota for entries with NO project (project_name=None)
    cost_no_project = sqlite_backend.get_accounting_entries_for_quota(
        start_time=start_time, limit_type=LimitType.COST, project_name=None, filter_project_null=True
    )
    assert cost_no_project == 0.5

def test_insert_and_get_usage_limits(sqlite_backend: SQLiteBackend, now_utc: datetime):
    limit1_created_at = (now_utc - timedelta(days=1)).replace(tzinfo=None)
    limit1_updated_at = (now_utc - timedelta(hours=12)).replace(tzinfo=None)
    
    limit_to_insert1 = UsageLimitDTO(
        scope=LimitScope.USER.value,
        limit_type=LimitType.COST.value,
        max_value=100.0,
        interval_unit=TimeInterval.MONTH.value,
        interval_value=1,
        username="test_user_1",
        created_at=limit1_created_at,
        updated_at=limit1_updated_at
    )
    sqlite_backend.insert_usage_limit(limit_to_insert1)

    limit2_created_at = (now_utc - timedelta(days=2)).replace(tzinfo=None)
    limit2_updated_at = (now_utc - timedelta(days=1)).replace(tzinfo=None)
    limit_to_insert2 = UsageLimitDTO(
        scope=LimitScope.MODEL.value,
        limit_type=LimitType.REQUESTS.value,
        max_value=1000.0,
        interval_unit=TimeInterval.DAY.value,
        interval_value=1,
        model="gpt-4-turbo",
        created_at=limit2_created_at,
        updated_at=limit2_updated_at
    )
    sqlite_backend.insert_usage_limit(limit_to_insert2)

    retrieved_limits: List[UsageLimitDTO] = sqlite_backend.get_usage_limits()
    assert len(retrieved_limits) == 2

    found_limit1 = False
    found_limit2 = False

    for limit_obj in retrieved_limits: # Renamed limit to limit_obj to avoid conflict
        assert isinstance(limit_obj, UsageLimitDTO)
        assert limit_obj.id is not None
        
        assert isinstance(limit_obj.created_at, datetime)
        assert isinstance(limit_obj.updated_at, datetime)
        assert limit_obj.created_at is not None and limit_obj.created_at.tzinfo == timezone.utc # Now expecting UTC aware
        assert limit_obj.updated_at is not None and limit_obj.updated_at.tzinfo == timezone.utc # Now expecting UTC aware

        if limit_obj.username == "test_user_1":
            found_limit1 = True
            assert limit_obj.scope == limit_to_insert1.scope
            assert limit_obj.limit_type == limit_to_insert1.limit_type
            assert limit_obj.max_value == limit_to_insert1.max_value
            assert limit_obj.interval_unit == limit_to_insert1.interval_unit
            assert limit_obj.interval_value == limit_to_insert1.interval_value
            # Compare aware datetime with aware datetime
            assert limit_obj.created_at is not None and limit_obj.created_at == limit1_created_at.replace(tzinfo=timezone.utc)
            assert limit_obj.updated_at is not None and limit_obj.updated_at == limit1_updated_at.replace(tzinfo=timezone.utc)
        elif limit_obj.model == "gpt-4-turbo":
            found_limit2 = True
            assert limit_obj.scope == limit_to_insert2.scope
            assert limit_obj.limit_type == limit_to_insert2.limit_type
            assert limit_obj.created_at is not None and limit_obj.created_at == limit2_created_at.replace(tzinfo=timezone.utc)
            assert limit_obj.updated_at is not None and limit_obj.updated_at == limit2_updated_at.replace(tzinfo=timezone.utc)
            
    assert found_limit1
    assert found_limit2

def test_get_usage_limits_with_filters(sqlite_backend: SQLiteBackend, now_utc: datetime):
    dt_naive = now_utc.replace(tzinfo=None) # For insertion
    dt_aware = now_utc # For comparison after retrieval

    limit1 = UsageLimitDTO(scope=LimitScope.USER.value, limit_type=LimitType.COST.value, max_value=100, interval_unit="month", interval_value=1, username="user1", created_at=dt_naive, updated_at=dt_naive)
    limit2 = UsageLimitDTO(scope=LimitScope.MODEL.value, limit_type=LimitType.REQUESTS.value, max_value=1000, interval_unit="day", interval_value=1, model="modelA", created_at=dt_naive, updated_at=dt_naive)
    limit3 = UsageLimitDTO(scope=LimitScope.USER.value, limit_type=LimitType.COST.value, max_value=200, interval_unit="month", interval_value=1, username="user2", model="modelA", created_at=dt_naive, updated_at=dt_naive)
    
    sqlite_backend.insert_usage_limit(limit1)
    sqlite_backend.insert_usage_limit(limit2)
    sqlite_backend.insert_usage_limit(limit3)

    user_limits = sqlite_backend.get_usage_limits(scope=LimitScope.USER)
    assert len(user_limits) == 2
    assert all(l.scope == LimitScope.USER.value for l in user_limits)

    model_a_limits = sqlite_backend.get_usage_limits(model="modelA")
    assert len(model_a_limits) == 2
    assert all(l.model == "modelA" for l in model_a_limits)

    user1_limits = sqlite_backend.get_usage_limits(username="user1")
    assert len(user1_limits) == 1
    assert user1_limits[0].username == "user1"

    user_model_limits = sqlite_backend.get_usage_limits(scope=LimitScope.USER, model="modelA")
    assert len(user_model_limits) == 1
    assert user_model_limits[0].username == "user2"
    assert user_model_limits[0].model == "modelA"


def test_delete_usage_limit(sqlite_backend: SQLiteBackend, now_utc: datetime):
    dt_naive = now_utc.replace(tzinfo=None)
    limit_to_delete_spec = UsageLimitDTO(
        scope=LimitScope.GLOBAL.value,
        limit_type=LimitType.COST.value,
        max_value=50.0,
        interval_unit=TimeInterval.WEEK.value,
        interval_value=1,
        caller_name="test_caller_delete",
        created_at=dt_naive,
        updated_at=dt_naive
    )
    sqlite_backend.insert_usage_limit(limit_to_delete_spec)

    all_limits = sqlite_backend.get_usage_limits(caller_name="test_caller_delete")
    assert len(all_limits) == 1
    limit_id_to_delete = all_limits[0].id
    assert limit_id_to_delete is not None

    sqlite_backend.delete_usage_limit(limit_id_to_delete)

    remaining_limits = sqlite_backend.get_usage_limits(caller_name="test_caller_delete")
    assert len(remaining_limits) == 0

    try:
        sqlite_backend.delete_usage_limit(99999)
    except Exception as e:
        pytest.fail(f"Deleting a non-existent limit raised an exception: {e}")

def test_datetime_precision_and_timezone_handling(sqlite_backend: SQLiteBackend):
    aware_dt = datetime.now(timezone.utc).replace(microsecond=123456)
    limit_aware = UsageLimitDTO(
        scope=LimitScope.GLOBAL.value, limit_type=LimitType.COST.value, max_value=1, interval_unit="day", interval_value=1,
        created_at=aware_dt, updated_at=aware_dt
    )
    sqlite_backend.insert_usage_limit(limit_aware)
    
    retrieved_aware_list = sqlite_backend.get_usage_limits(scope=LimitScope.GLOBAL)
    assert len(retrieved_aware_list) >= 1
    
    retrieved_aware = None
    for l_aware in retrieved_aware_list:
        if l_aware.created_at and l_aware.created_at == aware_dt: # Direct comparison for aware datetimes
             retrieved_aware = l_aware
             break
    assert retrieved_aware is not None, "Inserted aware_dt limit not found"
    
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.tzinfo == timezone.utc # Check it's aware
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.year == aware_dt.year
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.month == aware_dt.month
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.day == aware_dt.day
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.hour == aware_dt.hour
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.minute == aware_dt.minute
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.second == aware_dt.second
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.microsecond == aware_dt.microsecond
    assert retrieved_aware.created_at is not None and retrieved_aware.created_at.utcoffset() == timedelta(0)

    # Test with naive datetime (conventionally UTC)
    naive_dt = datetime.now(timezone.utc).replace(microsecond=654321)
    limit_naive = UsageLimitDTO(
        scope=LimitScope.USER.value, limit_type=LimitType.REQUESTS.value, max_value=1, interval_unit="hour", interval_value=1, username="naive_user",
        created_at=naive_dt, updated_at=naive_dt # Inserted as naive
    )
    sqlite_backend.insert_usage_limit(limit_naive)
    retrieved_naive_list = sqlite_backend.get_usage_limits(scope=LimitScope.USER, username="naive_user")
    assert len(retrieved_naive_list) == 1
    retrieved_naive_obj = retrieved_naive_list[0]

    # Expect retrieved datetime to be UTC-aware
    assert retrieved_naive_obj.created_at is not None and retrieved_naive_obj.created_at.tzinfo == timezone.utc
    # Compare by making the original naive_dt UTC-aware
    assert retrieved_naive_obj.created_at is not None and retrieved_naive_obj.created_at == naive_dt.replace(tzinfo=timezone.utc)
    assert retrieved_naive_obj.updated_at is not None and retrieved_naive_obj.updated_at.tzinfo == timezone.utc
    assert retrieved_naive_obj.updated_at is not None and retrieved_naive_obj.updated_at == naive_dt.replace(tzinfo=timezone.utc)


    # Test with None datetimes (should use DB defaults and be retrieved as UTC-aware)
    limit_none_dt = UsageLimitDTO(
        scope=LimitScope.CALLER.value, limit_type=LimitType.REQUESTS.value, max_value=10000, interval_unit="week", interval_value=1, caller_name="none_dt_caller",
        created_at=None, updated_at=None
    )
    sqlite_backend.insert_usage_limit(limit_none_dt)
    retrieved_none_dt_list = sqlite_backend.get_usage_limits(scope=LimitScope.CALLER, caller_name="none_dt_caller")
    assert len(retrieved_none_dt_list) == 1
    retrieved_none_dt = retrieved_none_dt_list[0]
    
    assert isinstance(retrieved_none_dt.created_at, datetime)
    assert retrieved_none_dt.created_at is not None and retrieved_none_dt.created_at.tzinfo == timezone.utc # Expect UTC aware from DB default
    assert isinstance(retrieved_none_dt.updated_at, datetime)
    assert retrieved_none_dt.updated_at is not None and retrieved_none_dt.updated_at.tzinfo == timezone.utc # Expect UTC aware from DB default
    
    current_utc_aware = datetime.now(timezone.utc)
    assert (current_utc_aware - retrieved_none_dt.created_at).total_seconds() < 10
    assert (current_utc_aware - retrieved_none_dt.updated_at).total_seconds() < 10



import os
import sys
from datetime import datetime
from io import StringIO
from unittest.mock import MagicMock, patch, call

import pytest

from llm_accounting import LLMAccounting, UsageEntry # Added UsageEntry
from llm_accounting.cli.main import main as cli_main


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage(mock_get_accounting):
    """Test tracking a new usage entry"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-3.5-turbo", "--prompt-tokens", "100", "--completion-tokens", "200", "--total-tokens", "300", "--cost", "0.02", "--execution-time", "0.5"]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()
    # Check that project is None by default
    args, _ = mock_backend_instance.insert_usage.call_args
    assert args[0].project is None


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage_with_project(mock_get_accounting):
    """Test tracking a new usage entry with a project"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance
    project_name = "MyAwesomeProject"

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-4", "--prompt-tokens", "10", "--cost", "0.01", "--execution-time", "0.1", "--project", project_name]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()
    # Check that project is passed correctly
    args, _ = mock_backend_instance.insert_usage.call_args
    assert isinstance(args[0], UsageEntry)
    assert args[0].project == project_name
    assert args[0].model == "gpt-4" # Ensure other args are still passed


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage_with_timestamp(mock_get_accounting):
    """Test tracking a new usage entry with a specific timestamp"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    timestamp = datetime(2023, 10, 1, 12, 0, 0)

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-3.5-turbo", "--prompt-tokens", "100", "--completion-tokens", "200", "--total-tokens", "300", "--cost", "0.02", "--execution-time", "0.5", "--timestamp", timestamp.strftime("%Y-%m-%d %H:%M:%S")]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage_with_caller_name(mock_get_accounting):
    """Test tracking a new usage entry with a caller name"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-3.5-turbo", "--prompt-tokens", "100", "--completion-tokens", "200", "--total-tokens", "300", "--cost", "0.02", "--execution-time", "0.5", "--caller-name", "test_app"]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage_with_username(mock_get_accounting):
    """Test tracking a new usage entry with a username"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-3.5-turbo", "--prompt-tokens", "100", "--completion-tokens", "200", "--total-tokens", "300", "--cost", "0.02", "--execution-time", "0.5", "--username", "test_user"]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage_with_cached_tokens(mock_get_accounting):
    """Test tracking a new usage entry with cached tokens"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-3.5-turbo", "--prompt-tokens", "100", "--completion-tokens", "200", "--total-tokens", "300", "--cost", "0.02", "--execution-time", "0.5", "--cached-tokens", "50"]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()


@patch("llm_accounting.cli.utils.get_accounting")
def test_track_usage_with_reasoning_tokens(mock_get_accounting):
    """Test tracking a new usage entry with reasoning tokens"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', "track", "--model", "gpt-3.5-turbo", "--prompt-tokens", "100", "--completion-tokens", "200", "--total-tokens", "300", "--cost", "0.02", "--execution-time", "0.5", "--reasoning-tokens", "50"]):
        cli_main()

    mock_backend_instance.insert_usage.assert_called_once()



import pytest
from datetime import datetime, timedelta, timezone
from llm_accounting import LLMAccounting, UsageEntry
from llm_accounting.models.limits import LimitScope, LimitType, TimeInterval, UsageLimit
from llm_accounting.services.quota_service import QuotaService
from llm_accounting.backends.sqlite import SQLiteBackend # Using real SQLite backend for these tests

@pytest.fixture
def sqlite_backend_for_quota(tmp_path):
    """Provides a clean SQLiteBackend instance for each test."""
    db_path = tmp_path / "test_quota_project.sqlite"
    backend = SQLiteBackend(db_path=str(db_path))
    with LLMAccounting(backend=backend) as acc: # Ensures tables are created
        acc.purge() # Clean slate
    return backend

@pytest.fixture
def accounting_for_quota(sqlite_backend_for_quota):
    """Provides an LLMAccounting instance with a clean SQLite backend."""
    with LLMAccounting(backend=sqlite_backend_for_quota) as acc:
        yield acc # acc.purge() is done by the backend fixture setup

# --- Helper to add usage ---
def add_usage(accounting_instance, model, cost, input_tokens, project_name=None, username="test_user", caller_name="test_caller", count=1):
    for _ in range(count):
        accounting_instance.track_usage(
            model=model,
            cost=cost,
            prompt_tokens=input_tokens, # Assuming input_tokens is representative for token limits
            project=project_name,
            username=username,
            caller_name=caller_name,
            timestamp=datetime.now(timezone.utc) # Ensure distinct timestamps if needed, or rely on ID
        )

# --- Project Scope Limit Tests ---

def test_project_limit_cost(accounting_for_quota: LLMAccounting):
    """Test cost limit scoped to a specific project."""
    project_a = "ProjectA"
    accounting_for_quota.set_usage_limit(
        scope=LimitScope.PROJECT,
        limit_type=LimitType.COST,
        max_value=5.0,
        interval_unit=TimeInterval.DAY,
        interval_value=1,
        project_name=project_a
    )

    # Usage for ProjectA
    add_usage(accounting_for_quota, model="gpt-4", cost=2.0, input_tokens=10, project_name=project_a, count=2) # Total cost 4.0
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=1.0, input_tokens=10, project_name=project_a, username="user1", caller_name="app1")
    assert allowed, "Should be allowed, under project limit"

    add_usage(accounting_for_quota, model="gpt-4", cost=1.0, input_tokens=10, project_name=project_a) # Total cost 5.0
    allowed, message = accounting_for_quota.check_quota(model="gpt-4", cost=0.01, input_tokens=1, project_name=project_a, username="user1", caller_name="app1")
    assert not allowed, "Should be denied, project limit reached"
    assert message is not None
    assert f"PROJECT (project: {project_a}) limit: 5.00 cost per 1 day, current usage: 5.00, request: 0.01" in message

    # Usage for ProjectB (should not be affected by ProjectA's limit)
    project_b = "ProjectB"
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=3.0, input_tokens=10, project_name=project_b, username="user1", caller_name="app1")
    assert allowed, "ProjectB usage should be allowed, not subject to ProjectA's limit"
    add_usage(accounting_for_quota, model="gpt-4", cost=3.0, input_tokens=10, project_name=project_b)


def test_project_limit_requests(accounting_for_quota: LLMAccounting):
    """Test request limit scoped to a specific project."""
    project_c = "ProjectC"
    accounting_for_quota.set_usage_limit(
        scope=LimitScope.PROJECT,
        limit_type=LimitType.REQUESTS,
        max_value=2.0, # Max 2 requests
        interval_unit=TimeInterval.DAY,
        interval_value=1,
        project_name=project_c
    )

    add_usage(accounting_for_quota, model="claude-2", cost=0.1, input_tokens=5, project_name=project_c, count=1)
    allowed, _ = accounting_for_quota.check_quota(model="claude-2", cost=0.1, input_tokens=5, project_name=project_c, username="user2", caller_name="app2")
    assert allowed, "Request 2 for ProjectC should be allowed"
    add_usage(accounting_for_quota, model="claude-2", cost=0.1, input_tokens=5, project_name=project_c, count=1) # Total 2 requests

    allowed, message = accounting_for_quota.check_quota(model="claude-2", cost=0.1, input_tokens=5, project_name=project_c, username="user2", caller_name="app2")
    assert not allowed, "Request 3 for ProjectC should be denied"
    assert message is not None
    assert f"PROJECT (project: {project_c}) limit: 2.00 requests per 1 day, current usage: 2.00, request: 1.00" in message

# --- Interaction with Global Limits ---

def test_project_limit_with_global_limit_cost(accounting_for_quota: LLMAccounting):
    """Test interaction between a project-specific cost limit and a global cost limit."""
    project_d = "ProjectD"
    # Global limit: 10.0 cost
    accounting_for_quota.set_usage_limit(LimitScope.GLOBAL, LimitType.COST, 10.0, TimeInterval.DAY, 1)
    # ProjectD limit: 5.0 cost
    accounting_for_quota.set_usage_limit(LimitScope.PROJECT, LimitType.COST, 5.0, TimeInterval.DAY, 1, project_name=project_d)

    # Use 4.0 cost for ProjectD (allowed by both project and global)
    add_usage(accounting_for_quota, model="gpt-4", cost=2.0, input_tokens=10, project_name=project_d, count=2)
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=1.0, input_tokens=10, project_name=project_d, username="user1", caller_name="app1")
    assert allowed, "ProjectD should be allowed (cost 4.0 + 1.0 = 5.0, project limit)"

    add_usage(accounting_for_quota, model="gpt-4", cost=1.0, input_tokens=10, project_name=project_d) # ProjectD total cost = 5.0

    # Next request for ProjectD should be denied by ProjectD's limit
    allowed, message = accounting_for_quota.check_quota(model="gpt-4", cost=0.1, input_tokens=1, project_name=project_d, username="user1", caller_name="app1")
    assert not allowed, "ProjectD should be denied by its own limit"
    assert message is not None
    assert f"PROJECT (project: {project_d}) limit: 5.00 cost per 1 day, current usage: 5.00, request: 0.10" in message

    # Use 6.0 cost for ProjectE (ProjectD is at 5.0, Global is at 5.0 + this 6.0 = 11.0)
    # This should be denied by the Global limit
    project_e = "ProjectE"
    add_usage(accounting_for_quota, model="gpt-4", cost=3.0, input_tokens=10, project_name=project_e, count=1) # ProjectE cost 3.0, Global 5.0+3.0=8.0
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=1.0, input_tokens=10, project_name=project_e, username="user1", caller_name="app1")
    assert allowed # ProjectE cost 4.0, Global 5.0+4.0=9.0

    add_usage(accounting_for_quota, model="gpt-4", cost=1.0, input_tokens=10, project_name=project_e) # ProjectE cost 4.0

    allowed, message = accounting_for_quota.check_quota(model="gpt-4", cost=1.1, input_tokens=10, project_name=project_e, username="user1", caller_name="app1")
    # ProjectD=5.0, ProjectE=4.0. Global usage = 9.0. Requesting 1.1 for ProjectE.
    # Total global would be 10.1, exceeding global limit of 10.0
    assert not allowed, "ProjectE should be denied by the global limit"
    assert message is not None
    assert f"GLOBAL limit: 10.00 cost per 1 day, current usage: 9.00, request: 1.10" in message


def test_project_limit_with_model_limit(accounting_for_quota: LLMAccounting):
    """Test interaction between project limit and model limit."""
    project_f = "ProjectF"
    model_name = "special-model"

    # Model limit for "special-model": 3 requests
    accounting_for_quota.set_usage_limit(LimitScope.MODEL, LimitType.REQUESTS, 3, TimeInterval.DAY, 1, model=model_name)
    # ProjectF limit for "special-model": 2 requests
    accounting_for_quota.set_usage_limit(LimitScope.PROJECT, LimitType.REQUESTS, 2, TimeInterval.DAY, 1, model=model_name, project_name=project_f)

    # Request 1 for special-model in ProjectF (allowed)
    add_usage(accounting_for_quota, model=model_name, cost=0.1, input_tokens=1, project_name=project_f)
    allowed, _ = accounting_for_quota.check_quota(model=model_name, cost=0.1, input_tokens=1, project_name=project_f, username="u", caller_name="c")
    assert allowed

    # Request 2 for special-model in ProjectF (allowed by project, hits project limit)
    add_usage(accounting_for_quota, model=model_name, cost=0.1, input_tokens=1, project_name=project_f)
    allowed, message = accounting_for_quota.check_quota(model=model_name, cost=0.1, input_tokens=1, project_name=project_f, username="u", caller_name="c")
    assert not allowed, "Should be denied by ProjectF limit"
    assert message is not None
    assert f"PROJECT (model: {model_name}, project: {project_f}) limit: 2.00 requests per 1 day, current usage: 2.00, request: 1.00" in message

    # Request for special-model in ProjectG (ProjectF limit doesn't apply)
    project_g = "ProjectG"
    # This is the 3rd request for "special-model" overall.
    add_usage(accounting_for_quota, model=model_name, cost=0.1, input_tokens=1, project_name=project_g)
    allowed, message = accounting_for_quota.check_quota(model=model_name, cost=0.1, input_tokens=1, project_name=project_g, username="u", caller_name="c")
    assert not allowed, "3rd request for special-model (in ProjectG) should be denied by model limit"
    assert message is not None
    assert f"MODEL (model: {model_name}) limit: 3.00 requests per 1 day, current usage: 3.00, request: 1.00" in message

def test_project_limit_with_no_specific_project_in_request(accounting_for_quota: LLMAccounting):
    """Test that a request with no project is not affected by a project-specific limit."""
    project_h = "ProjectH"
    accounting_for_quota.set_usage_limit(
        scope=LimitScope.PROJECT,
        limit_type=LimitType.COST,
        max_value=1.0,
        interval_unit=TimeInterval.DAY,
        interval_value=1,
        project_name=project_h
    )
    # This request has no project, so ProjectH's limit should not apply.
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=2.0, input_tokens=10, project_name=None, username="user1", caller_name="app1")
    assert allowed, "Request with no project should not be affected by ProjectH's limit"

    # Add usage to ProjectH to hit its limit
    add_usage(accounting_for_quota, model="gpt-4", cost=1.0, input_tokens=10, project_name=project_h)
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=0.1, input_tokens=1, project_name=project_h, username="user1", caller_name="app1")
    assert not allowed, "ProjectH should now be over limit"

    # Another request with no project should still be allowed
    allowed, _ = accounting_for_quota.check_quota(model="gpt-4", cost=2.0, input_tokens=10, project_name=None, username="user1", caller_name="app1")
    assert allowed, "Request with no project should still be allowed even if ProjectH is over limit"

def test_limit_message_for_project_scope(accounting_for_quota: LLMAccounting):
    """Test the limit exhaustion message for project-scoped limits."""
    project_name = "DetailProject"
    model_name = "detailed-model"
    accounting_for_quota.set_usage_limit(
        scope=LimitScope.PROJECT,
        limit_type=LimitType.REQUESTS,
        max_value=1,
        interval_unit=TimeInterval.DAY,
        interval_value=1,
        project_name=project_name,
        model=model_name # Also associate with a model for more specific message
    )
    add_usage(accounting_for_quota, model=model_name, cost=0.1, input_tokens=1, project_name=project_name)
    allowed, message = accounting_for_quota.check_quota(model=model_name, cost=0.1, input_tokens=1, project_name=project_name, username="u", caller_name="c")
    
    assert not allowed
    assert message is not None
    assert f"PROJECT (model: {model_name}, project: {project_name}) limit: 1.00 requests per 1 day, current usage: 1.00, request: 1.00" in message



from datetime import datetime, timedelta, timezone
from unittest.mock import MagicMock, call # Added call
import pytest
from typing import Optional # Added Optional

# Updated import: UsageLimit changed to UsageLimitData
# Also importing enums needed for creating UsageLimitData instances
from llm_accounting.models.limits import (LimitScope, LimitType, TimeInterval, # Added TimeInterval
                                          UsageLimitDTO)
from llm_accounting.services.quota_service import QuotaService
from llm_accounting.backends.base import BaseBackend # For type hinting the mock_backend


# Fixture for a mock backend, as QuotaService depends on a backend instance
@pytest.fixture
def mock_backend() -> MagicMock:
    """Provides a MagicMock instance for BaseBackend."""
    backend = MagicMock(spec=BaseBackend)
    return backend

@pytest.fixture
def quota_service(mock_backend: MagicMock) -> QuotaService:
    """Provides a QuotaService instance initialized with a mock backend."""
    return QuotaService(mock_backend)


def test_check_quota_no_limits(quota_service: QuotaService, mock_backend: MagicMock):
    """Test check_quota when no limits are configured."""
    mock_backend.get_usage_limits.return_value = []
    
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=100, cost=0.01
    )
    
    assert is_allowed is True
    assert reason is None
    # Check that get_usage_limits was called multiple times (e.g., 5 times for the different scopes)
    # and that for each call, the appropriate parameters were used.
    # The exact number of calls might change if QuotaService logic changes.
    # Based on current QuotaService structure: MODEL, GLOBAL, USER, CALLER (general), CALLER (user+caller)
    assert mock_backend.get_usage_limits.call_count >= 1 # At least one call for GLOBAL if others are specific
    
    expected_calls = [
        call(scope=LimitScope.MODEL, model="gpt-4"),
        call(scope=LimitScope.GLOBAL),
        call(scope=LimitScope.USER, username="test_user"),
        call(scope=LimitScope.CALLER, caller_name="test_caller"), # General caller
        call(scope=LimitScope.CALLER, username="test_user", caller_name="test_caller") # User+Caller
    ]
    # Use any_order=True because the internal order of fetching these specific scopes might not be guaranteed
    # or strictly necessary for this test's purpose if all return empty.
    # However, QuotaService has a defined order of checks. Let's assume that for "no limits",
    # it will try all relevant specific scopes.
    # A simpler check might be to just ensure call_count is as expected (e.g., 5)
    # and that it's called with variations of the input parameters.
    # For this test, the main point is that it eventually returns True because all calls yield no limits.
    # So, just checking call_count might be sufficient if we trust other tests to verify specific call args.
    assert mock_backend.get_usage_limits.call_count == 5 # Based on the 5 checks in QuotaService


def test_check_quota_allowed_single_limit(quota_service: QuotaService, mock_backend: MagicMock):
    """Test check_quota when usage is within a single configured limit."""
    now = datetime.now(timezone.utc)
    limit = UsageLimitDTO(
        id=1, scope=LimitScope.USER.value, limit_type=LimitType.COST.value,
        max_value=10.0, interval_unit=TimeInterval.MONTH.value, interval_value=1,
        username="test_user", created_at=now, updated_at=now
    )
    mock_backend.get_usage_limits.return_value = [limit]
    # Current usage for the period (e.g., $5 for the month)
    mock_backend.get_accounting_entries_for_quota.return_value = 5.0 
    
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=100, cost=0.01 # Request cost is $0.01
    )
    
    assert is_allowed is True
    assert reason is None

    # Define a side_effect function for mock_backend.get_usage_limits
    # This function will be called each time mock_backend.get_usage_limits is invoked
    def get_usage_limits_side_effect(scope: Optional[LimitScope] = None, model: Optional[str] = None, 
                                     username: Optional[str] = None, caller_name: Optional[str] = None, 
                                     project_name: Optional[str] = None, filter_project_null: Optional[bool] = None,
                                     filter_username_null: Optional[bool] = None, filter_caller_name_null: Optional[bool] = None):
        # Only return the 'limit' if the scope and username match the specific USER limit we're testing
        if scope == LimitScope.USER and username == "test_user":
            return [limit] 
        return [] # For all other calls (MODEL, GLOBAL, CALLER scopes), return no limits

    mock_backend.get_usage_limits.side_effect = get_usage_limits_side_effect
    
    # Reset mocks if they were called during initial setup of side_effect (though not strictly necessary here)
    # mock_backend.get_usage_limits.reset_mock() # Not needed as it's fresh for this test logic
    mock_backend.get_accounting_entries_for_quota.reset_mock() # Reset this as it's called by check_quota

    # Re-run check_quota with the new side_effect in place for get_usage_limits
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=100, cost=0.01 # Request cost is $0.01
    )
    
    assert is_allowed is True # Should still be allowed
    assert reason is None

    # Verify that get_usage_limits was called for the USER scope with correct parameters
    # (among other calls QuotaService makes)
    mock_backend.get_usage_limits.assert_any_call(scope=LimitScope.USER, username="test_user")
    
    # Verify that get_accounting_entries_for_quota was called ONCE, 
    # specifically for the USER limit that was found and evaluated.
    mock_backend.get_accounting_entries_for_quota.assert_called_once()
    # Correctly access keyword arguments
    kwargs = mock_backend.get_accounting_entries_for_quota.call_args.kwargs
    assert kwargs['limit_type'] == LimitType.COST # limit_type argument from the 'limit' object
    # The username passed to get_accounting_entries_for_quota should be the one from the limit object
    # if the limit is user-specific.
    assert kwargs['username'] == "test_user"    # username argument (from the limit's scope)

def test_check_quota_denied_single_limit(quota_service: QuotaService, mock_backend: MagicMock):
    """Test check_quota when usage exceeds a single configured limit."""
    now = datetime.now(timezone.utc)
    limit = UsageLimitDTO(
        id=1, scope=LimitScope.USER.value, limit_type=LimitType.COST.value,
        max_value=10.0, interval_unit=TimeInterval.MONTH.value, interval_value=1,
        username="test_user", created_at=now, updated_at=now
    )
    mock_backend.get_usage_limits.return_value = [limit]
    # Current usage for the period (e.g., $9.99 for the month)
    mock_backend.get_accounting_entries_for_quota.return_value = 9.99
    
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=0, cost=0.02 # Request cost is $0.02, total would be $10.01
    )
    
    assert is_allowed is False
    assert reason is not None
    
    # Define a side_effect function for mock_backend.get_usage_limits
    def get_usage_limits_side_effect_denied(scope: Optional[LimitScope] = None, model: Optional[str] = None, 
                                            username: Optional[str] = None, caller_name: Optional[str] = None, project_name: Optional[str] = None):
        if scope == LimitScope.USER and username == "test_user":
            return [limit] 
        return []

    mock_backend.get_usage_limits.side_effect = get_usage_limits_side_effect_denied # Use unique name
    # Reset mocks before the second call to check_quota
    mock_backend.get_accounting_entries_for_quota.reset_mock()
    mock_backend.get_usage_limits.reset_mock() # Also reset this one
    mock_backend.get_accounting_entries_for_quota.return_value = 9.99 # Ensure it's set for the re-run

    # Re-run check_quota
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=0, cost=0.02 
    )
    assert is_allowed is False # Should still be false
    assert reason is not None

    # Corrected expected message: QuotaService uses the 'limit_scope_for_message' passed to _evaluate_limits.
    # _check_user_limits passes "USER".
    assert "USER (user: test_user) limit: 10.00 cost per 1 month" in reason
    assert "current usage: 9.99, request: 0.02" in reason

    # Verify calls
    mock_backend.get_usage_limits.assert_any_call(scope=LimitScope.USER, username="test_user")
    mock_backend.get_accounting_entries_for_quota.assert_called_once()


def test_check_quota_multiple_limits_one_exceeded(quota_service: QuotaService, mock_backend: MagicMock):
    """Test check_quota with multiple limits, where one is exceeded."""
    now = datetime.now(timezone.utc)
    cost_limit = UsageLimitDTO(
        id=1, scope=LimitScope.USER.value, limit_type=LimitType.COST.value,
        max_value=10.0, interval_unit=TimeInterval.MONTH.value, interval_value=1,
        username="test_user", created_at=now, updated_at=now
    )
    request_limit = UsageLimitDTO(
        id=2, scope=LimitScope.USER.value, limit_type=LimitType.REQUESTS.value,
        max_value=100.0, interval_unit=TimeInterval.DAY.value, interval_value=1,
        username="test_user", created_at=now, updated_at=now
    )
    mock_backend.get_usage_limits.return_value = [cost_limit, request_limit]

    # Scenario: Cost is fine, but requests are exceeded
    def get_accounting_side_effect(start_time, limit_type, model, username, caller_name, project_name, filter_project_null):
        if limit_type == LimitType.COST:
            return 5.0 # Well within $10 cost limit
        elif limit_type == LimitType.REQUESTS:
            return 100.0 # Already at 100 requests, next one (count as 1) will exceed
        return 0.0
    
    mock_backend.get_accounting_entries_for_quota.side_effect = get_accounting_side_effect
    
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=10, cost=0.01 # Request cost is $0.01, request count is 1
    )
    
    assert is_allowed is False
    assert reason is not None

    # Define a side_effect for get_usage_limits
    def get_usage_limits_side_effect_multiple(scope: Optional[LimitScope] = None, model: Optional[str] = None, 
                                              username: Optional[str] = None, caller_name: Optional[str] = None, project_name: Optional[str] = None):
        if scope == LimitScope.USER and username == "test_user":
            return [cost_limit, request_limit] # Return both user-specific limits
        return []

    mock_backend.get_usage_limits.side_effect = get_usage_limits_side_effect_multiple # Corrected name
    
    # Reset get_accounting_entries_for_quota mock and set its new side_effect for this test
    mock_backend.get_accounting_entries_for_quota.reset_mock()
    mock_backend.get_accounting_entries_for_quota.side_effect = get_accounting_side_effect

    # Re-run check_quota
    is_allowed, reason = quota_service.check_quota(
        model="gpt-4", username="test_user", caller_name="test_caller",
        input_tokens=10, cost=0.01 
    )
    assert is_allowed is False # Should still be false
    assert reason is not None

    # The failure message will be for the first limit that's hit (REQUESTS in this case).
    # Expected scope message is "USER" as these are user limits.
    assert "USER (user: test_user) limit: 100.00 requests per 1 day" in reason
    assert "current usage: 100.00, request: 1.00" in reason # Corrected float formatting

    # Verify calls
    mock_backend.get_usage_limits.assert_any_call(scope=LimitScope.USER, username="test_user")
    # get_accounting_entries_for_quota will be called for COST limit (passes) and then for REQUESTS limit (fails)
    assert mock_backend.get_accounting_entries_for_quota.call_count == 2


def test_check_quota_different_scopes(quota_service: QuotaService, mock_backend: MagicMock):
    """Test that limits are fetched with combined scopes if applicable."""
    # QuotaService currently fetches limits by ORing scope attributes in get_usage_limits call.
    # The mock_backend.get_usage_limits is called once with all relevant user, model, caller.
    # So, this test mainly ensures the call to get_usage_limits is made as expected.
    
    mock_backend.get_usage_limits.return_value = [] # No limits for simplicity
    
    quota_service.check_quota(
        model="super_model", username="super_user", caller_name="super_caller",
        input_tokens=1, cost=0.001
    )
    
    # Based on the QuotaService logic, it will make multiple calls for different scopes.
    # The specific parameters for each call depend on how QuotaService constructs them.
    # MODEL: model="super_model"
    # GLOBAL: (no specific entity filters)
    # USER: username="super_user"
    # CALLER (general): caller_name="super_caller", username=None
    # CALLER (user+caller): username="super_user", caller_name="super_caller"

    # Corrected order: MODEL, USER, CALLER (general), CALLER (user+caller), GLOBAL
    # Parameters not specified in the call from QuotaService will default to None.
    
    # Define expected calls as a list of dictionaries for manual comparison
    expected_calls_data = [
        {"scope": LimitScope.MODEL, "model": "super_model", "username": None, "caller_name": None},
        {"scope": LimitScope.USER, "username": "super_user", "model": None, "caller_name": None},
        {"scope": LimitScope.CALLER, "caller_name": "super_caller", "username": None, "model": None}, # General caller limit
        {"scope": LimitScope.CALLER, "username": "super_user", "caller_name": "super_caller", "model": None},  # User + Caller specific limit
        {"scope": LimitScope.GLOBAL, "model": None, "username": None, "caller_name": None} # GLOBAL is checked last
    ]

    actual_calls = mock_backend.get_usage_limits.call_args_list
    assert len(actual_calls) == len(expected_calls_data), \
        f"Expected {len(expected_calls_data)} calls, but got {len(actual_calls)}. Actual calls: {actual_calls}"

    actual_calls_kwargs_list = [c.kwargs for c in actual_calls]

    for expected_kwargs_item in expected_calls_data:
        found_match = False
        # Try to find a match in a way that respects that actual_kwargs_item might have more keys (if defaults are added by mock)
        # but all keys in expected_kwargs_item must match.
        for actual_kwargs_item in actual_calls_kwargs_list:
            match = True
            if not (actual_kwargs_item.get('scope') == expected_kwargs_item['scope'] and
                    actual_kwargs_item.get('model') == expected_kwargs_item.get('model') and
                    actual_kwargs_item.get('username') == expected_kwargs_item.get('username') and
                    actual_kwargs_item.get('caller_name') == expected_kwargs_item.get('caller_name') and
                    actual_kwargs_item.get('project_name') == expected_kwargs_item.get('project_name')):
                match = False
            
            if match:
                found_match = True
                # To ensure each actual call is matched at most once (like assert_has_calls with any_order=True)
                # we can remove it from a temporary list. For simplicity here, if distinct expected calls, this is okay.
                # If expected calls could be identical, a more complex removal strategy is needed.
                # For this test, expected_calls_data are distinct.
                break 
        assert found_match, f"Expected call with kwargs {expected_kwargs_item} not found in actual calls. Actual calls: {actual_calls_kwargs_list}"
    
    # Also verify the total number of calls if it's strictly 5 (already covered by len assertion)
    assert mock_backend.get_usage_limits.call_count == 5


def test_check_quota_token_limits(quota_service: QuotaService, mock_backend: MagicMock):
    """Test check_quota for input token limits."""
    now = datetime.now(timezone.utc)
    token_limit = UsageLimitDTO(
        id=1, scope=LimitScope.MODEL.value, limit_type=LimitType.INPUT_TOKENS.value,
        max_value=1000.0, interval_unit=TimeInterval.HOUR.value, interval_value=1,
        model="text-davinci-003", created_at=now, updated_at=now
    )
    mock_backend.get_usage_limits.return_value = [token_limit]
    mock_backend.get_accounting_entries_for_quota.return_value = 950.0 # 950 tokens used this hour

    # Scenario 1: Allowed
    is_allowed, reason = quota_service.check_quota(
        model="text-davinci-003", username="any_user", caller_name="any_caller",
        input_tokens=50, cost=0.0 # 50 tokens requested
    )
    assert is_allowed is True
    assert reason is None

    # Scenario 2: Denied
    is_allowed, reason = quota_service.check_quota(
        model="text-davinci-003", username="any_user", caller_name="any_caller",
        input_tokens=51, cost=0.0 # 51 tokens requested, total 1001, limit 1000
    )
    assert is_allowed is False
    assert reason is not None

    # Define a side_effect for get_usage_limits
    def get_usage_limits_side_effect_model(scope: Optional[LimitScope] = None, model: Optional[str] = None, 
                                           username: Optional[str] = None, caller_name: Optional[str] = None, project_name: Optional[str] = None):
        if scope == LimitScope.MODEL and model == "text-davinci-003":
            return [token_limit] # Return the model-specific limit
        return []

    # Reset mocks and apply new side_effect for get_usage_limits
    # mock_backend.get_usage_limits.reset_mock() # Not strictly needed if calls are specific enough
    mock_backend.get_usage_limits.side_effect = get_usage_limits_side_effect_model
    mock_backend.get_accounting_entries_for_quota.reset_mock()
    mock_backend.get_accounting_entries_for_quota.return_value = 950.0 # Set for this scenario

    # Re-run the denied scenario check
    is_allowed, reason = quota_service.check_quota(
        model="text-davinci-003", username="any_user", caller_name="any_caller",
        input_tokens=51, cost=0.0 
    )
    assert is_allowed is False # Should still be false
    assert reason is not None

    # This is a MODEL limit, so the scope in message should be MODEL
    assert "MODEL (model: text-davinci-003) limit: 1000.00 input_tokens per 1 hour" in reason
    assert "current usage: 950.00, request: 51.00" in reason # Values are formatted to .2f

    # Verify calls
    mock_backend.get_usage_limits.assert_any_call(scope=LimitScope.MODEL, model="text-davinci-003")
    mock_backend.get_accounting_entries_for_quota.assert_called_once()


def test_get_period_start_monthly(quota_service: QuotaService):
    """Test _get_period_start for monthly interval."""
    # Test for a specific date
    current_time = datetime(2024, 3, 15, 10, 30, 0, tzinfo=timezone.utc)
    period_start = quota_service._get_period_start(current_time, TimeInterval.MONTH, 1)
    assert period_start == datetime(2024, 3, 1, 0, 0, 0, tzinfo=timezone.utc)

    # Test for beginning of month
    current_time = datetime(2024, 4, 1, 0, 0, 0, tzinfo=timezone.utc)
    period_start = quota_service._get_period_start(current_time, TimeInterval.MONTH, 1)
    assert period_start == datetime(2024, 4, 1, 0, 0, 0, tzinfo=timezone.utc)

def test_get_period_start_daily(quota_service: QuotaService):
    """Test _get_period_start for daily interval."""
    current_time = datetime(2024, 3, 15, 10, 30, 0, tzinfo=timezone.utc)
    period_start = quota_service._get_period_start(current_time, TimeInterval.DAY, 1)
    assert period_start == datetime(2024, 3, 15, 0, 0, 0, tzinfo=timezone.utc)

def test_get_period_start_hourly(quota_service: QuotaService):
    """Test _get_period_start for hourly interval."""
    current_time = datetime(2024, 3, 15, 10, 30, 0, tzinfo=timezone.utc)
    period_start = quota_service._get_period_start(current_time, TimeInterval.HOUR, 1)
    assert period_start == datetime(2024, 3, 15, 10, 0, 0, tzinfo=timezone.utc)

# More tests could be added for other intervals (WEEK, MINUTE, SECOND) and different interval_value
# and for cases where TimeInterval.MONTH.value is "monthly" string vs TimeInterval.MONTH enum.
# The QuotaService._get_period_start seems to handle enum directly.
# The UsageLimitData stores interval_unit as string. QuotaService should handle this.
# In check_quota, limit.interval_unit (string) is converted to TimeInterval enum.
# This seems fine.



import re
import os

def get_current_version(file_path="pyproject.toml"):
    with open(file_path, "r") as f:
        content = f.read()
    version_match = re.search(r'version = "(\d+\.\d+\.\d+)"', content)
    if not version_match:
        raise ValueError("Version string not found in pyproject.toml")
    return version_match.group(1)

def increment_version(file_path="pyproject.toml"):
    with open(file_path, "r") as f:
        content = f.read()

    version_match = re.search(r'version = "(\d+\.\d+\.\d+)"', content)
    if not version_match:
        raise ValueError("Version string not found in pyproject.toml")

    old_version = version_match.group(1)
    version_parts = list(map(int, old_version.split('.')))
    version_parts[-1] += 1  # Increment the last part (revision)
    new_version = ".".join(map(str, version_parts))

    new_content = re.sub(r'version = "\d+\.\d+\.\d+"', f'version = "{new_version}"', content)

    with open(file_path, "w") as f:
        f.write(new_content)
    return new_version

def revert_version(original_version, file_path="pyproject.toml"):
    with open(file_path, "r") as f:
        content = f.read()
    new_content = re.sub(r'version = "\d+\.\d+\.\d+"', f'version = "{original_version}"', content)
    with open(file_path, "w") as f:
        f.write(new_content)
    print(f"Reverted version to {original_version}")

if __name__ == "__main__":
    # This part is for testing or direct use, not for the orchestrator
    # The orchestrator will call these functions directly.
    pass



import subprocess
import sys
import os
from version_manager import get_current_version, increment_version, revert_version

def run_command(command, shell=True, check=True):
    """Helper to run shell commands."""
    print(f"Executing: {command}")
    try:
        result = subprocess.run(command, shell=shell, check=check, capture_output=True, text=True)
        print(result.stdout)
        if result.stderr:
            print(result.stderr, file=sys.stderr)
        return result
    except subprocess.CalledProcessError as e:
        print(f"Command failed with error: {e}", file=sys.stderr)
        print(f"Stdout: {e.stdout}", file=sys.stderr)
        print(f"Stderr: {e.stderr}", file=sys.stderr)
        raise

def main():
    original_version = None
    try:
        # 1. Get original version
        original_version = get_current_version()
        print(f"Current version: {original_version}")

        # 2. Increment version in pyproject.toml
        new_version = increment_version()
        print(f"Incremented version to: {new_version}")

        # 3. Clean up previous build artifacts
        print("Cleaning up build artifacts...")
        run_command("rmdir /s /q dist", check=False) # check=False because rmdir might fail if dir doesn't exist

        # 4. Build the package
        print("Building package...")
        run_command("python -m build")

        # 5. Upload to PyPI
        print("Uploading to PyPI...")
        # Assuming PYPI_TOKEN is set as an environment variable
        pypi_token = os.environ.get("PYPI_TOKEN")
        if not pypi_token:
            print("PYPI_TOKEN environment variable not set. Skipping PyPI upload.", file=sys.stderr)
            # Decide whether to exit or continue without upload
            sys.exit(1) # Exit if token is missing for upload

        run_command(f'twine upload --username __token__ --password "{pypi_token}" dist/*')

        # 6. Create GitHub Release
        print("Creating GitHub Release...")
        run_command(f'gh release create {new_version} --title "Release {new_version}" --notes "Automated release for version {new_version}"')

        print("Release process completed successfully!")

    except Exception as e:
        print(f"An error occurred during the release process: {e}", file=sys.stderr)
        if original_version:
            print("Attempting to revert version...", file=sys.stderr)
            try:
                revert_version(original_version)
                print("Version reverted successfully.", file=sys.stderr)
            except Exception as revert_e:
                print(f"Failed to revert version: {revert_e}", file=sys.stderr)
        sys.exit(1) # Exit with error code

if __name__ == "__main__":
    # Ensure virtual environment is activated before running this script
    # This script assumes it's run within an activated venv or that python -m build etc. are globally available
    # For build.bat, we'll activate the venv there.
    main()



import logging

logger = logging.getLogger(__name__)

class SchemaManager:
    def __init__(self, backend_instance):
        self.backend = backend_instance

    # _create_schema_if_not_exists and _create_tables methods are removed.
    # DDL operations are now handled by SQLAlchemy's Base.metadata.create_all
    # (called from PostgreSQLBackend.initialize) and eventually Alembic for migrations.
    # This class is kept for potential future schema-related utility methods
    # that do not perform DDL directly (e.g., schema validation checks via SQLAlchemy inspect).
    # For example, a method like `check_tables_exist(self, table_names: list[str]) -> bool`
    # could be added here if needed elsewhere, using `inspect(self.backend.engine)`.



import sys
from io import StringIO
from unittest.mock import MagicMock, patch

import pytest

from llm_accounting import LLMAccounting
from llm_accounting.cli.main import main as cli_main


@patch("llm_accounting.cli.utils.get_accounting")
def test_select_non_select_query(mock_get_accounting, test_db, capsys):
    """Test rejection of non-SELECT queries"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance
    mock_backend_instance.execute_query.side_effect = ValueError("Only SELECT queries are allowed")

    with patch.object(sys, 'argv', ['cli_main', "select", "--query", "INSERT INTO accounting_entries (model) VALUES ('gpt-4')"]):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            cli_main()

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1
    captured = capsys.readouterr()
    assert "Error executing query: Only SELECT queries are allowed" in captured.out



from rich.table import Table

from llm_accounting import LLMAccounting

from ..utils import console, format_float, format_time, format_tokens


def run_tail(args, accounting: LLMAccounting):
    """Show the most recent usage entries"""
    entries = accounting.tail(args.number)

    if not entries:
        console.print("[yellow]No usage entries found[/yellow]")
        return

    # Create table for entries
    table = Table(title=f"Last {len(entries)} Usage Entries")
    table.add_column("Timestamp", style="cyan")
    table.add_column("Model", style="cyan")
    table.add_column("Project", style="cyan")
    table.add_column("Caller", style="cyan")
    table.add_column("User", style="cyan")
    table.add_column("Prompt Tokens", justify="right", style="green")
    table.add_column("Completion Tokens", justify="right", style="green")
    table.add_column("Total Tokens", justify="right", style="green")
    table.add_column("Cost", justify="right", style="green")
    table.add_column("Exec Time", justify="right", style="green")

    for entry in entries:
        table.add_row(
            entry.timestamp.strftime("%Y-%m-%d %H:%M:%S") if entry.timestamp else "-",
            entry.model,
            entry.project or "-",
            entry.caller_name or "-",
            entry.username or "-",
            format_tokens(
                entry.prompt_tokens if entry.prompt_tokens is not None else 0
            ),
            format_tokens(
                entry.completion_tokens if entry.completion_tokens is not None else 0
            ),
            format_tokens(entry.total_tokens if entry.total_tokens is not None else 0),
            f"${format_float(entry.cost)}",
            format_time(entry.execution_time),
        )

    console.print(table)



import sys
from datetime import datetime
from unittest.mock import patch, MagicMock

from llm_accounting import LLMAccounting
from llm_accounting.cli.main import main as cli_main
from llm_accounting.cli.utils import format_tokens, format_float, format_time # For verifying formatted output


def make_entry(**kwargs):
    entry = MagicMock()
    entry.model = kwargs.get("model", "gpt-4")
    entry.prompt_tokens = kwargs.get("prompt_tokens", 0)
    entry.completion_tokens = kwargs.get("completion_tokens", 0)
    entry.total_tokens = kwargs.get("total_tokens", 0)
    entry.cost = kwargs.get("cost", 0.0)
    entry.execution_time = kwargs.get("execution_time", 0.0)
    entry.timestamp = kwargs.get("timestamp", datetime(2024, 1, 1, 12, 0, 0))
    entry.caller_name = kwargs.get("caller_name", "")
    entry.username = kwargs.get("username", "")
    entry.project = kwargs.get("project", None)
    return entry


@patch("llm_accounting.cli.utils.get_accounting")
def test_tail_default(mock_get_accounting, capsys):
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance
    
    entry1_data = {"model":"gpt-4", "prompt_tokens":100, "completion_tokens":50, "total_tokens":150, "cost":0.002, "execution_time":1.5, "caller_name":"test_application_long_name", "username":"test_user_long_name", "project":None}
    entry2_data = {"model":"gpt-3.5-turbo", "prompt_tokens":200, "completion_tokens":100, "total_tokens":300, "cost":0.003, "execution_time":2.0, "project":"TestProjectLongName"} # Made project name longer
    
    mock_backend_instance.tail.return_value = [
        make_entry(**entry1_data),
        make_entry(**entry2_data)
    ]

    with patch.object(sys, 'argv', ['cli_main', "tail"]):
        cli_main()
    captured = capsys.readouterr().out
    
    assert "Last 2 Usage Entries" in captured
    
    # Check for entry1 data
    assert "gpt-4" in captured # gpt-4 is short enough
    assert "test…" in captured 
    assert format_tokens(entry1_data["prompt_tokens"]) in captured
    assert format_tokens(entry1_data["completion_tokens"]) in captured
    assert format_tokens(entry1_data["total_tokens"]) in captured
    assert "$0.0…" in captured 
    assert "1.5…" in captured 
    assert "│ -         │" in captured or "│ -          │" in captured or "│ -     │" in captured # Project is None

    # Check for entry2 data
    assert "gpt-…" in captured # gpt-3.5-turbo will be truncated
    assert format_tokens(entry2_data["prompt_tokens"]) in captured
    assert format_tokens(entry2_data["completion_tokens"]) in captured
    assert format_tokens(entry2_data["total_tokens"]) in captured
    assert "$0.0…" in captured 
    assert "2.0…" in captured 
    assert "Test…" in captured # Truncated project name "TestProjectLongName"


@patch("llm_accounting.cli.utils.get_accounting")
def test_tail_custom_number(mock_get_accounting, capsys):
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance
    
    entry_data = {"model":"gpt-4", "prompt_tokens":100, "completion_tokens":50, "total_tokens":150, "cost":0.002, "execution_time":1.5, "caller_name":"test_application_long_name", "username":"test_user_long_name", "project":"MyProject"}
    mock_backend_instance.tail.return_value = [
        make_entry(**entry_data)
    ]

    with patch.object(sys, 'argv', ['cli_main', "tail", "-n", "5"]):
        cli_main()
    captured = capsys.readouterr().out
    
    assert "Last 1 Usage Entry" in captured or "Last 1 Usage Entries" in captured
    assert "gpt-4" in captured
    assert "test…" in captured 
    assert format_tokens(entry_data["prompt_tokens"]) in captured
    assert format_tokens(entry_data["completion_tokens"]) in captured
    assert format_tokens(entry_data["total_tokens"]) in captured
    assert "$0.0…" in captured 
    assert "1.5…" in captured 
    assert "MyPr…" in captured or "MyProject" in captured # Handle potential truncation for MyProject


@patch("llm_accounting.cli.utils.get_accounting")
def test_tail_empty(mock_get_accounting, capsys):
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance
    mock_backend_instance.tail.return_value = []

    with patch.object(sys, 'argv', ['cli_main', "tail"]):
        cli_main()
    captured = capsys.readouterr()
    assert "No usage entries found" in captured.out



from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple
from typing_extensions import override


from llm_accounting.backends.base import BaseBackend, UsageEntry, UsageStats
from llm_accounting.models.limits import LimitScope, LimitType, UsageLimitDTO


class MockBackend(BaseBackend):
    """A mock backend that implements all required methods for BaseBackend interface testing."""

    @override
    def _ensure_connected(self) -> None:
        """Mocks ensuring connection."""
        pass

    @override
    def initialize(self) -> None:
        pass

    @override
    def insert_usage(self, entry: UsageEntry) -> None:
        pass

    @override
    def get_period_stats(self, start: datetime, end: datetime) -> UsageStats:
        return UsageStats()

    @override
    def get_model_stats(self, start: datetime, end: datetime) -> List[Tuple[str, UsageStats]]:
        return []

    @override
    def get_model_rankings(self, start: datetime, end: datetime) -> Dict[str, List[Tuple[str, Any]]]:
        return {}

    @override
    def purge(self) -> None:
        pass

    @override
    def tail(self, n: int = 10) -> List[UsageEntry]:
        return []

    @override
    def close(self) -> None:
        pass

    @override
    def execute_query(self, query: str) -> List[Dict[str, Any]]:
        """Mock implementation of execute_query"""
        if query.strip().upper().startswith("SELECT"):
            return [{}] 
        raise ValueError("MockBackend in test_base.py only supports SELECT for execute_query.")


    @override
    def get_usage_limits(
        self,
        scope: Optional[LimitScope] = None,
        model: Optional[str] = None,
        username: Optional[str] = None,
        caller_name: Optional[str] = None,
        project_name: Optional[str] = None,
        filter_project_null: Optional[bool] = None,
        filter_username_null: Optional[bool] = None,
        filter_caller_name_null: Optional[bool] = None,
    ) -> List[UsageLimitDTO]:
        return []

    @override
    def get_accounting_entries_for_quota(
        self,
        start_time: datetime,
        limit_type: LimitType,
        model: Optional[str] = None,
        username: Optional[str] = None,
        caller_name: Optional[str] = None,
        project_name: Optional[str] = None,
        filter_project_null: Optional[bool] = None # New parameter
    ) -> float:
        return 0.0

    @override
    def insert_usage_limit(self, limit: UsageLimitDTO) -> None:
        pass

    @override
    def delete_usage_limit(self, limit_id: int) -> None:
        pass

    @override
    def initialize_audit_log_schema(self) -> None:
        pass

    @override
    def log_audit_event(self, entry: Any) -> None: # Using Any for entry to avoid circular import for AuditLogEntry
        pass

    @override
    def get_audit_log_entries(self, start_date: Optional[datetime] = None, end_date: Optional[datetime] = None, app_name: Optional[str] = None, user_name: Optional[str] = None, project: Optional[str] = None, log_type: Optional[str] = None, limit: Optional[int] = None) -> List[Any]: # Using Any for AuditLogEntry
        return []


class IncompleteBackend(BaseBackend):
    """A mock backend that doesn't implement all required methods for BaseBackend interface testing."""

    def initialize(self) -> None:
        pass

    @override
    def _ensure_connected(self) -> None: pass
    @override
    def insert_usage(self, entry: UsageEntry) -> None: pass
    @override
    def get_period_stats(self, start: datetime, end: datetime) -> UsageStats: return UsageStats()
    @override
    def get_model_stats(self, start: datetime, end: datetime) -> List[Tuple[str, UsageStats]]: return []
    @override
    def get_model_rankings(self, start: datetime, end: datetime) -> Dict[str, List[Tuple[str, Any]]]: return {}
    @override
    def purge(self) -> None: pass
    @override
    def tail(self, n: int = 10) -> List[UsageEntry]: return []
    @override
    def close(self) -> None: pass
    @override
    def execute_query(self, query: str) -> List[Dict[str, Any]]: return [{}]
    @override
    def get_usage_limits(self, scope: Optional[LimitScope] = None, model: Optional[str] = None, username: Optional[str] = None, caller_name: Optional[str] = None, project_name: Optional[str] = None, filter_project_null: Optional[bool] = None, filter_username_null: Optional[bool] = None, filter_caller_name_null: Optional[bool] = None) -> List[UsageLimitDTO]: return []
    @override
    def get_accounting_entries_for_quota(self, start_time: datetime, limit_type: LimitType, model: Optional[str] = None, username: Optional[str] = None, caller_name: Optional[str] = None, project_name: Optional[str] = None, filter_project_null: Optional[bool] = None) -> float: return 0.0
    @override
    def insert_usage_limit(self, limit: UsageLimitDTO) -> None: pass

    @override
    def initialize_audit_log_schema(self) -> None: pass

    @override
    def log_audit_event(self, entry: Any) -> None: pass

    @override
    def get_audit_log_entries(self, start_date: Optional[datetime] = None, end_date: Optional[datetime] = None, app_name: Optional[str] = None, user_name: Optional[str] = None, project: Optional[str] = None, log_type: Optional[str] = None, limit: Optional[int] = None) -> List[Any]: return []



from datetime import datetime
from typing import Dict, List, Tuple
from sqlalchemy import text
from sqlalchemy.engine import Connection # For type hinting

from llm_accounting.backends.base import UsageEntry, UsageStats


def insert_usage_query(conn: Connection, entry: UsageEntry) -> None:
    """Insert a new usage entry into the database using named parameters."""
    params = {
        "timestamp": entry.timestamp.isoformat() if entry.timestamp else None,
        "model": entry.model,
        "prompt_tokens": entry.prompt_tokens,
        "completion_tokens": entry.completion_tokens,
        "total_tokens": entry.total_tokens,
        "local_prompt_tokens": entry.local_prompt_tokens,
        "local_completion_tokens": entry.local_completion_tokens,
        "local_total_tokens": entry.local_total_tokens,
        "cost": entry.cost,
        "execution_time": entry.execution_time,
        "caller_name": entry.caller_name,
        "username": entry.username,
        "cached_tokens": entry.cached_tokens,
        "reasoning_tokens": entry.reasoning_tokens,
        "project": entry.project,
    }
    
    sql = text("""
        INSERT INTO accounting_entries (
            timestamp, model, prompt_tokens, completion_tokens, total_tokens,
            local_prompt_tokens, local_completion_tokens, local_total_tokens,
            cost, execution_time, caller_name, username, cached_tokens, reasoning_tokens, project
        ) VALUES (
            :timestamp, :model, :prompt_tokens, :completion_tokens, :total_tokens,
            :local_prompt_tokens, :local_completion_tokens, :local_total_tokens,
            :cost, :execution_time, :caller_name, :username, :cached_tokens, :reasoning_tokens, :project
        )
    """)
    conn.execute(sql, params)
    # Removed conn.commit() - let the caller in SQLiteBackend handle transaction management.


def get_period_stats_query(
    conn: Connection, start: datetime, end: datetime
) -> UsageStats:
    """Get aggregated statistics for a time period from the database using named parameters."""
    sql = text("""
        SELECT
            SUM(prompt_tokens) as sum_prompt_tokens,
            SUM(completion_tokens) as sum_completion_tokens,
            SUM(total_tokens) as sum_total_tokens,
            SUM(local_prompt_tokens) as sum_local_prompt_tokens,
            SUM(local_completion_tokens) as sum_local_completion_tokens,
            SUM(local_total_tokens) as sum_local_total_tokens,
            SUM(cost) as sum_cost,
            SUM(execution_time) as sum_execution_time,
            AVG(prompt_tokens) as avg_prompt_tokens,
            AVG(completion_tokens) as avg_completion_tokens,
            AVG(total_tokens) as avg_total_tokens,
            AVG(local_prompt_tokens) as avg_local_prompt_tokens,
            AVG(local_completion_tokens) as avg_local_completion_tokens,
            AVG(local_total_tokens) as avg_local_total_tokens,
            AVG(cost) as avg_cost,
            AVG(execution_time) as avg_execution_time
        FROM accounting_entries
        WHERE timestamp BETWEEN :start_time AND :end_time
    """)
    
    result = conn.execute(sql, {"start_time": start.isoformat(), "end_time": end.isoformat()})
    row = result.fetchone()

    if not row or row.sum_cost is None: # Check if any aggregation happened (e.g. sum_cost is a good indicator)
        return UsageStats(
            sum_prompt_tokens=0, sum_completion_tokens=0, sum_total_tokens=0,
            sum_local_prompt_tokens=0, sum_local_completion_tokens=0, sum_local_total_tokens=0,
            sum_cost=0.0, sum_execution_time=0.0,
            avg_prompt_tokens=0.0, avg_completion_tokens=0.0, avg_total_tokens=0.0,
            avg_local_prompt_tokens=0.0, avg_local_completion_tokens=0.0, avg_local_total_tokens=0.0,
            avg_cost=0.0, avg_execution_time=0.0,
        )

    return UsageStats(
        sum_prompt_tokens=row.sum_prompt_tokens or 0,
        sum_completion_tokens=row.sum_completion_tokens or 0,
        sum_total_tokens=row.sum_total_tokens or 0,
        sum_local_prompt_tokens=row.sum_local_prompt_tokens or 0,
        sum_local_completion_tokens=row.sum_local_completion_tokens or 0,
        sum_local_total_tokens=row.sum_local_total_tokens or 0,
        sum_cost=row.sum_cost or 0.0,
        sum_execution_time=row.sum_execution_time or 0.0,
        avg_prompt_tokens=row.avg_prompt_tokens or 0.0,
        avg_completion_tokens=row.avg_completion_tokens or 0.0,
        avg_total_tokens=row.avg_total_tokens or 0.0,
        avg_local_prompt_tokens=row.avg_local_prompt_tokens or 0.0,
        avg_local_completion_tokens=row.avg_local_completion_tokens or 0.0,
        avg_local_total_tokens=row.avg_local_total_tokens or 0.0,
        avg_cost=row.avg_cost or 0.0,
        avg_execution_time=row.avg_execution_time or 0.0,
    )


def get_model_stats_query(
    conn: Connection, start: datetime, end: datetime
) -> List[Tuple[str, UsageStats]]:
    """Get statistics grouped by model for a time period from the database using named parameters."""
    sql = text("""
        SELECT
            model,
            SUM(prompt_tokens) as sum_prompt_tokens,
            SUM(completion_tokens) as sum_completion_tokens,
            SUM(total_tokens) as sum_total_tokens,
            SUM(local_prompt_tokens) as sum_local_prompt_tokens,
            SUM(local_completion_tokens) as sum_local_completion_tokens,
            SUM(local_total_tokens) as sum_local_total_tokens,
            SUM(cost) as sum_cost,
            SUM(execution_time) as sum_execution_time,
            AVG(prompt_tokens) as avg_prompt_tokens,
            AVG(completion_tokens) as avg_completion_tokens,
            AVG(total_tokens) as avg_total_tokens,
            AVG(local_prompt_tokens) as avg_local_prompt_tokens,
            AVG(local_completion_tokens) as avg_local_completion_tokens,
            AVG(local_total_tokens) as avg_local_total_tokens,
            AVG(cost) as avg_cost,
            AVG(execution_time) as avg_execution_time
        FROM accounting_entries
        WHERE timestamp BETWEEN :start_time AND :end_time
        GROUP BY model
    """)
    
    result = conn.execute(sql, {"start_time": start.isoformat(), "end_time": end.isoformat()})
    rows = result.fetchall()

    return [
        (
            str(row.model), # Ensure model is string
            UsageStats(
                sum_prompt_tokens=row.sum_prompt_tokens or 0,
                sum_completion_tokens=row.sum_completion_tokens or 0,
                sum_total_tokens=row.sum_total_tokens or 0,
                sum_local_prompt_tokens=row.sum_local_prompt_tokens or 0,
                sum_local_completion_tokens=row.sum_local_completion_tokens or 0,
                sum_local_total_tokens=row.sum_local_total_tokens or 0,
                sum_cost=row.sum_cost or 0.0,
                sum_execution_time=row.sum_execution_time or 0.0,
                avg_prompt_tokens=row.avg_prompt_tokens or 0.0,
                avg_completion_tokens=row.avg_completion_tokens or 0.0,
                avg_total_tokens=row.avg_total_tokens or 0.0,
                avg_local_prompt_tokens=row.avg_local_prompt_tokens or 0.0,
                avg_local_completion_tokens=row.avg_local_completion_tokens or 0.0,
                avg_local_total_tokens=row.avg_local_total_tokens or 0.0,
                avg_cost=row.avg_cost or 0.0,
                avg_execution_time=row.avg_execution_time or 0.0,
            ),
        )
        for row in rows
    ]


def get_model_rankings_query(
    conn: Connection, start: datetime, end: datetime
) -> Dict[str, List[Tuple[str, float]]]:
    """Get model rankings based on different metrics from the database using named parameters."""
    params = {"start_time": start.isoformat(), "end_time": end.isoformat()}

    prompt_tokens_sql = text("""
        SELECT model, SUM(prompt_tokens) as total
        FROM accounting_entries
        WHERE timestamp BETWEEN :start_time AND :end_time
        GROUP BY model
        ORDER BY total DESC
    """)
    result_prompt = conn.execute(prompt_tokens_sql, params)
    prompt_tokens_ranking = [(str(row.model), float(row.total or 0)) for row in result_prompt.fetchall()]

    cost_sql = text("""
        SELECT model, SUM(cost) as total
        FROM accounting_entries
        WHERE timestamp BETWEEN :start_time AND :end_time
        GROUP BY model
        ORDER BY total DESC
    """)
    result_cost = conn.execute(cost_sql, params)
    cost_ranking = [(str(row.model), float(row.total or 0.0)) for row in result_cost.fetchall()]

    return {"prompt_tokens": prompt_tokens_ranking, "cost": cost_ranking}


def tail_query(conn: Connection, n: int = 10) -> List[UsageEntry]:
    """Get the n most recent usage entries from the database using named parameters."""
    sql = text("""
        SELECT
            timestamp, model, prompt_tokens, completion_tokens, total_tokens,
            local_prompt_tokens, local_completion_tokens, local_total_tokens,
            cost, execution_time, caller_name, username, cached_tokens, reasoning_tokens, project
        FROM accounting_entries
        ORDER BY timestamp DESC
        LIMIT :limit_n
    """)
    
    result = conn.execute(sql, {"limit_n": n})
    rows = result.fetchall()

    return [
        UsageEntry(
            model=row.model,
            prompt_tokens=row.prompt_tokens,
            completion_tokens=row.completion_tokens,
            total_tokens=row.total_tokens,
            local_prompt_tokens=row.local_prompt_tokens,
            local_completion_tokens=row.local_completion_tokens,
            local_total_tokens=row.local_total_tokens,
            cost=row.cost,
            execution_time=row.execution_time,
            timestamp=datetime.fromisoformat(row.timestamp), # Ensure timestamp is datetime object
            caller_name=row.caller_name,
            username=row.username,
            cached_tokens=row.cached_tokens,
            reasoning_tokens=row.reasoning_tokens,
            project=row.project,
        )
        for row in rows
    ]



import logging
import os
import psycopg2
import psycopg2.extras  # For RealDictCursor
import psycopg2.extensions  # For connection type
from typing import Optional, List, Tuple, Dict, Any
from datetime import datetime

from sqlalchemy import create_engine, text, inspect
from llm_accounting.models.base import Base

from .base import BaseBackend, UsageEntry, UsageStats, AuditLogEntry # Added AuditLogEntry
from ..models.limits import UsageLimitDTO, LimitScope, LimitType

from .postgresql_backend_parts.connection_manager import ConnectionManager
from .postgresql_backend_parts.schema_manager import SchemaManager
from .postgresql_backend_parts.data_inserter import DataInserter
from .postgresql_backend_parts.data_deleter import DataDeleter
from .postgresql_backend_parts.query_executor import QueryExecutor
from .postgresql_backend_parts.limit_manager import LimitManager  # Import LimitManager

logger = logging.getLogger(__name__)


class PostgreSQLBackend(BaseBackend):
    conn: Optional[psycopg2.extensions.connection] = None
    """
    A backend for llm-accounting that uses a PostgreSQL database, specifically
    tailored for Neon serverless Postgres but compatible with standard PostgreSQL instances.
    """

    def __init__(self, postgresql_connection_string: Optional[str] = None):
        """
        Initializes the PostgreSQLBackend.
        """
        if postgresql_connection_string:
            self.connection_string = postgresql_connection_string
        else:
            self.connection_string = os.environ.get("POSTGRESQL_CONNECTION_STRING")

        if not self.connection_string:
            raise ValueError(
                "PostgreSQL connection string not provided and POSTGRESQL_CONNECTION_STRING "
                "environment variable is not set."
            )
        self.conn = None # This will be the psycopg2 connection
        self.engine = None # This will be the SQLAlchemy engine
        logger.info("PostgreSQLBackend initialized with connection string.")

        self.connection_manager = ConnectionManager(self)
        self.schema_manager = SchemaManager(self) # SchemaManager no longer does DDL
        self.data_inserter = DataInserter(self)
        self.data_deleter = DataDeleter(self)
        self.query_executor = QueryExecutor(self)
        # Instantiate LimitManager, passing the backend instance and data_inserter instance
        self.limit_manager = LimitManager(self, self.data_inserter)

    def initialize(self) -> None:
        """
        Connects to the PostgreSQL database, sets up the SQLAlchemy engine,
        and ensures the schema is created via SQLAlchemy models if tables are missing.
        The psycopg2 connection is also initialized for existing components.
        """
        # Initialize psycopg2 connection via ConnectionManager (Option A)
        self.connection_manager.initialize() 
        logger.info("psycopg2 connection initialized by ConnectionManager.")

        # Create SQLAlchemy engine if it doesn't exist
        if not self.engine:
            if not self.connection_string: # Should have been caught in __init__ but double check
                raise ValueError("Cannot initialize SQLAlchemy engine: Connection string is missing.")
            try:
                self.engine = create_engine(self.connection_string, future=True)
                logger.info("SQLAlchemy engine created successfully.")
            except Exception as e:
                logger.error(f"Failed to create SQLAlchemy engine: {e}")
                raise

        # Inspect existing tables and create schema if necessary
        if self.engine:
            try:
                inspector = inspect(self.engine)
                existing_tables = inspector.get_table_names()
                
                # Get table names from SQLAlchemy metadata
                tables_to_create = []
                for table_obj in Base.metadata.sorted_tables:
                    # PostgreSQL stores table names in lowercase by default unless quoted.
                    # SQLAlchemy model names are typically CamelCase for classes, but table names are snake_case.
                    # The actual table name is taken from __tablename__.
                    if table_obj.name not in existing_tables:
                         tables_to_create.append(table_obj.name)

                if tables_to_create:
                    logger.info(f"New tables to create based on SQLAlchemy models: {tables_to_create}. Creating schema...")
                    Base.metadata.create_all(self.engine) # Creates only missing tables
                    logger.info("Schema creation/update from SQLAlchemy models complete.")
                else:
                    logger.info("All tables defined in SQLAlchemy models already exist. Schema creation via Base.metadata.create_all skipped (Alembic will handle migrations).")
            except Exception as e:
                logger.error(f"Error during schema inspection or creation with SQLAlchemy: {e}")
                # Depending on policy, might want to raise this or handle differently
                raise
        else:
            # This case should ideally not be reached if engine creation is successful
            logger.error("SQLAlchemy engine not available. Cannot perform schema check/creation.")
            raise RuntimeError("SQLAlchemy engine could not be initialized.")

        # The call to self.schema_manager._create_schema_if_not_exists() is removed.
        # self.schema_manager can be used for other checks if needed in the future.

    def close(self) -> None:
        """
        Closes the psycopg2 connection and disposes of the SQLAlchemy engine.
        """
        self.connection_manager.close() # Closes the psycopg2 connection
        if self.engine:
            logger.info("Disposing SQLAlchemy engine.")
            self.engine.dispose()
            self.engine = None

    # _create_schema_if_not_exists and _create_tables methods are removed from PostgreSQLBackend
    # as this responsibility is now with initialize() using SQLAlchemy and Alembic for migrations.

    def insert_usage(self, entry: UsageEntry) -> None:
        self.data_inserter.insert_usage(entry)

    def insert_usage_limit(self, limit: UsageLimitDTO) -> None:
        """
        Inserts a usage limit into the usage_limits table.
        Delegates to LimitManager.
        """
        self._ensure_connected()
        self.limit_manager.insert_usage_limit(limit)

    def delete_usage_limit(self, limit_id: int) -> None:
        self.data_deleter.delete_usage_limit(limit_id)

    def get_period_stats(self, start: datetime, end: datetime) -> UsageStats:
        return self.query_executor.get_period_stats(start, end)

    def get_model_stats(self, start: datetime, end: datetime) -> List[Tuple[str, UsageStats]]:
        return self.query_executor.get_model_stats(start, end)

    def get_model_rankings(self, start: datetime, end: datetime) -> Dict[str, List[Tuple[str, Any]]]:
        return self.query_executor.get_model_rankings(start, end)

    def tail(self, n: int = 10) -> List[UsageEntry]:
        return self.query_executor.tail(n)

    def purge(self) -> None:
        self.data_deleter.purge()

    def get_usage_limits(
            self,
            scope: Optional[LimitScope] = None,
            model: Optional[str] = None,
            username: Optional[str] = None,
            caller_name: Optional[str] = None,
            project_name: Optional[str] = None,
            filter_project_null: Optional[bool] = None,
            filter_username_null: Optional[bool] = None,
            filter_caller_name_null: Optional[bool] = None) -> List[UsageLimitDTO]:
        """
        Retrieves usage limits (as UsageLimitData objects) from the `usage_limits` table
        based on specified filter criteria. Delegates to LimitManager.
        """
        self._ensure_connected()
        return self.limit_manager.get_usage_limits(
            scope=scope,
            model=model,
            username=username,
            caller_name=caller_name,
            project_name=project_name,
            filter_project_null=filter_project_null,
            filter_username_null=filter_username_null,
            filter_caller_name_null=filter_caller_name_null
        )

    def get_accounting_entries_for_quota(
            self,
            start_time: datetime,
            limit_type: LimitType,
            model: Optional[str] = None,
            username: Optional[str] = None,
            caller_name: Optional[str] = None,
            project_name: Optional[str] = None,
            filter_project_null: Optional[bool] = None) -> float:
        self._ensure_connected()
        if self.conn is None:
            raise ConnectionError("Database connection is not established.")

        if limit_type == LimitType.REQUESTS:
            agg_field = "COUNT(*)"
        elif limit_type == LimitType.INPUT_TOKENS:
            agg_field = "COALESCE(SUM(prompt_tokens), 0)"
        elif limit_type == LimitType.OUTPUT_TOKENS:
            agg_field = "COALESCE(SUM(completion_tokens), 0)"
        elif limit_type == LimitType.COST:
            agg_field = "COALESCE(SUM(cost), 0.0)"
        else:
            logger.error(f"Unsupported LimitType for quota aggregation: {limit_type}")
            raise ValueError(f"Unsupported LimitType for quota aggregation: {limit_type}")

        base_query = f"SELECT {agg_field} AS aggregated_value FROM accounting_entries"
        conditions = ["timestamp >= %s"]
        params: List[Any] = [start_time]

        if model:
            conditions.append("model_name = %s")
            params.append(model)
        if username:
            conditions.append("username = %s")
            params.append(username)
        if caller_name:
            conditions.append("caller_name = %s")
            params.append(caller_name)
        
        if project_name is not None:
            conditions.append("project = %s")
            params.append(project_name)
        if filter_project_null is True:
            conditions.append("project IS NULL")
        if filter_project_null is False:
            conditions.append("project IS NOT NULL")

        query = base_query
        if conditions:
            query += " WHERE " + " AND ".join(conditions)
        query += ";"

        try:
            with self.conn.cursor() as cur:
                cur.execute(query, tuple(params))
                result = cur.fetchone()
                if result and result[0] is not None:
                    return float(result[0])
                return 0.0
        except psycopg2.Error as e:
            logger.error(f"Error getting accounting entries for quota (type: {limit_type.value}): {e}")
            raise
        except Exception as e:
            logger.error(f"An unexpected error occurred getting accounting entries for quota (type: {limit_type.value}): {e}")
            raise

    def execute_query(self, query: str) -> List[Dict[str, Any]]:
        self._ensure_connected()
        if self.conn is None:
            raise ConnectionError("Database connection is not established.")

        if not query.lstrip().upper().startswith("SELECT"):
            logger.error(f"Attempted to execute non-SELECT query: {query}")
            raise ValueError("Only SELECT queries are allowed for execution via this method.")
        results = []
        try:
            with self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor) as cur:
                cur.execute(query)
                results = [dict(row) for row in cur.fetchall()]
            logger.info(f"Successfully executed custom query. Rows returned: {len(results)}")
            return results
        except psycopg2.Error as e:
            logger.error(f"Error executing query '{query}': {e}")
            raise
        except Exception as e:
            logger.error(f"An unexpected error occurred executing query '{query}': {e}")
            raise

    def get_usage_costs(self, user_id: str, start_date: Optional[datetime] = None, end_date: Optional[datetime] = None) -> float:
        return self.query_executor.get_usage_costs(user_id, start_date, end_date)

    def set_usage_limit(
            self,
            user_id: str,
            limit_amount: float,
            limit_type_str: str = "COST") -> None:
        """
        A simplified way to set a usage limit for a user.
        Delegates to QueryExecutor.
        NOTE: This method is distinct from the BaseBackend's insert_usage_limit.
              It's a convenience method. The primary method for inserting limits
              is now `insert_usage_limit(self, limit_data: UsageLimitData)`.
              This one might need to be refactored or deprecated.
        """
        self.query_executor.set_usage_limit(user_id, limit_amount, limit_type_str)

    def get_usage_limit(self, user_id: str) -> Optional[List[UsageLimitDTO]]:
        """
        Retrieves all usage limits (as UsageLimitData) for a specific user.
        Delegates to LimitManager.
        NOTE: This method is distinct from BaseBackend's get_usage_limits.
              It's a convenience method.
        """
        self._ensure_connected()
        return self.limit_manager.get_usage_limit(user_id, project_name=None)

    def _ensure_connected(self) -> None:
        self.connection_manager.ensure_connected()

    # Audit Log Methods Implementation

    def initialize_audit_log_schema(self) -> None:
        """
        Ensures the audit log schema is initialized.
        In this backend, the main initialize() method handles all schema creation.
        This method primarily ensures the connection is active.
        """
        self.connection_manager.ensure_connected()
        # self.schema_manager._create_schema_if_not_exists() # This is called by initialize()
        logger.info("Audit log schema initialization check (delegated to main initialize).")

    def log_audit_event(self, entry: AuditLogEntry) -> None:
        """
        Logs an audit event to the database.
        Manages connection, transaction commit, and rollback.
        """
        self.connection_manager.ensure_connected()
        assert self.conn is not None, "Database connection is not established for logging audit event."

        try:
            self.data_inserter.insert_audit_log_event(entry)
            self.conn.commit()
            logger.info(f"Audit event logged successfully for user '{entry.user_name}', app '{entry.app_name}'.")
        except psycopg2.Error as e:
            logger.error(f"Database error logging audit event: {e}")
            if self.conn and not self.conn.closed:
                try:
                    self.conn.rollback()
                    logger.info("Transaction rolled back due to error logging audit event.")
                except psycopg2.Error as rb_err:
                    logger.error(f"Error during rollback attempt: {rb_err}")
            # Consider re-raising a more generic error or a specific application error
            raise RuntimeError(f"Failed to log audit event due to database error: {e}") from e
        except Exception as e:
            logger.error(f"Failed to log audit event due to database error: {e}")
            if self.conn and not self.conn.closed:
                try:
                    self.conn.rollback()
                    logger.info("Transaction rolled back due to error logging audit event.")
                except psycopg2.Error as rb_err:
                    logger.error(f"Error during rollback attempt: {rb_err}")
            raise RuntimeError(f"Failed to log audit event due to database error: {e}") from e

    def get_audit_log_entries(
        self,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        app_name: Optional[str] = None,
        user_name: Optional[str] = None,
        project: Optional[str] = None,
        log_type: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> List[AuditLogEntry]:
        """
        Retrieves audit log entries based on specified filter criteria.
        Delegates to QueryExecutor and manages connection.
        """
        self.connection_manager.ensure_connected()
        assert self.conn is not None, "Database connection is not established for retrieving audit log entries."

        try:
            entries = self.query_executor.get_audit_log_entries(
                start_date=start_date,
                end_date=end_date,
                app_name=app_name,
                user_name=user_name,
                project=project,
                log_type=log_type,
                limit=limit,
            )
            logger.info(f"Retrieved {len(entries)} audit log entries.")
            return entries
        except psycopg2.Error as e:
            logger.error(f"Database error retrieving audit log entries: {e}")
            # For read operations, rollback is typically not needed unless part of a larger transaction
            raise RuntimeError(f"Failed to retrieve audit log entries due to database error: {e}") from e
        except Exception as e:
            logger.error(f"Unexpected error retrieving audit log entries: {e}")
            raise RuntimeError(f"Unexpected error occurred while retrieving audit log entries: {e}") from e



import sys
from io import StringIO
from unittest.mock import MagicMock, patch

from llm_accounting import LLMAccounting
from llm_accounting.cli.main import main as cli_main


@patch("llm_accounting.cli.utils.get_accounting")
def test_purge_with_confirmation(mock_get_accounting):
    """Test purge command with confirmation by checking accounting calls"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    # Simulate user input 'y'
    with patch('sys.stdin', StringIO('y\n')), patch.object(sys, 'argv', ['cli_main', 'purge']):
        cli_main()

    mock_backend_instance.purge.assert_called_once()


@patch("llm_accounting.cli.utils.get_accounting")
def test_purge_without_confirmation(mock_get_accounting):
    """Test purge command without confirmation"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    # Simulate user input 'n'
    with patch('sys.stdin', StringIO('n\n')), patch.object(sys, 'argv', ['cli_main', 'purge']):
        cli_main()

    mock_backend_instance.purge.assert_not_called()


@patch("llm_accounting.cli.utils.get_accounting")
def test_purge_with_yes_flag(mock_get_accounting):
    """Test purge command with -y flag by checking accounting calls"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', 'purge', '-y']):
        cli_main()

    mock_backend_instance.purge.assert_called_once()


@patch("llm_accounting.cli.utils.get_accounting")
def test_purge_with_yes_flag_long(mock_get_accounting):
    """Test purge command with --yes flag by checking accounting calls"""
    mock_backend_instance = MagicMock()
    real_accounting_instance = LLMAccounting(backend=mock_backend_instance)
    mock_get_accounting.return_value = real_accounting_instance

    with patch.object(sys, 'argv', ['cli_main', 'purge', '--yes']):
        cli_main()

    mock_backend_instance.purge.assert_called_once()



import unittest
from unittest.mock import patch, MagicMock, call
import os
from datetime import datetime, timezone, timedelta

from src.llm_accounting.backends.postgresql import PostgreSQLBackend
from src.llm_accounting.backends.base import UsageEntry, UsageStats, AuditLogEntry
from src.llm_accounting.models.limits import UsageLimitDTO, LimitScope, LimitType, TimeInterval
from typing import List, Optional

import psycopg2

from tests.backends.postgresql_backend_tests.base_test_postgresql import BaseTestPostgreSQL


class TestPostgreSQLAuditLog(BaseTestPostgreSQL):

    def test_initialize_creates_audit_log_table(self):
        """
        Tests that initialize calls SchemaManager._create_schema_if_not_exists,
        which is responsible for all table creations including audit_log_entries.
        """
        self.backend.initialize() # This calls _create_schema_if_not_exists
        self.mock_schema_manager_instance._create_schema_if_not_exists.assert_called_once()
        # Further verification that _create_schema_if_not_exists itself calls _create_tables
        # which then includes the audit log DDL would be part of SchemaManager's own tests.
        # Here we confirm NeonBackend delegates overall schema setup.

    def test_initialize_audit_log_schema_method_ensures_connection(self):
        """
        Tests the specific initialize_audit_log_schema method.
        It should ensure connection. Schema creation is handled by the main initialize.
        """
        # Reset mocks that might have been called during self.backend instantiation or previous calls
        self.mock_psycopg2_module.connect.reset_mock()
        
        # Simulate connection not being active initially
        self.backend.conn = None
        
        self.backend.initialize_audit_log_schema()
        
        # Check if connection_manager.ensure_connected (which calls psycopg2.connect if conn is None) was triggered
        self.mock_psycopg2_module.connect.assert_called_once_with('dummy_dsn_from_env')
        self.assertEqual(self.backend.conn, self.mock_conn)
        # Also check that _create_schema_if_not_exists is NOT called by this specific method,
        # as it's handled by the main initialize().
        self.mock_schema_manager_instance._create_schema_if_not_exists.assert_not_called()


    def test_log_audit_event_delegates_and_manages_transaction(self):
        """
        Tests that log_audit_event delegates to DataInserter and manages transactions.
        """
        self.backend.initialize() # Ensure connection is set up
        
        sample_entry = AuditLogEntry(
            timestamp=datetime.now(timezone.utc),
            app_name="test_app",
            user_name="test_user",
            model="test_model",
            log_type="test_log",
            id=None, # id is managed by DB
            prompt_text=None, response_text=None, remote_completion_id=None, project=None
        )
        self.backend.log_audit_event(sample_entry)

        self.mock_data_inserter_instance.insert_audit_log_event.assert_called_once_with(sample_entry)
        self.mock_conn.commit.assert_called_once()
        self.mock_conn.rollback.assert_not_called()

        # Test rollback on DataInserter failure
        self.mock_data_inserter_instance.reset_mock()
        self.mock_conn.reset_mock()
        self.mock_data_inserter_instance.insert_audit_log_event.side_effect = self.mock_psycopg2_module.Error("DB error on insert")
        
        with self.assertRaisesRegex(RuntimeError, r"Failed to log audit event due to database error: DB error on insert"):
            self.backend.log_audit_event(sample_entry)
        
        self.mock_data_inserter_instance.insert_audit_log_event.assert_called_once_with(sample_entry)
        self.mock_conn.commit.assert_not_called()
        self.mock_conn.rollback.assert_called_once()

    def test_get_audit_log_entries_delegates_to_query_executor(self):
        """
        Tests that get_audit_log_entries delegates to QueryExecutor.
        """
        self.backend.initialize() # Ensure connection is set up

        now = datetime.now(timezone.utc)
        expected_entries_data = [
            AuditLogEntry(timestamp=now, app_name="app1", user_name="user1", model="model1", log_type="type1", id=1, prompt_text=None, response_text=None, remote_completion_id=None, project=None),
            AuditLogEntry(timestamp=now, app_name="app2", user_name="user2", model="model2", log_type="type2", id=2, prompt_text=None, response_text=None, remote_completion_id=None, project=None)
        ]
        self.mock_query_executor_instance.get_audit_log_entries.return_value = expected_entries_data

        filter_params = {
            "start_date": now - timedelta(days=1),
            "end_date": now,
            "app_name": "test_app_filter",
            "user_name": "test_user_filter",
            "project": "TestProject",
            "log_type": "test_type_filter",
            "limit": 50
        }
        
        retrieved_entries = self.backend.get_audit_log_entries(**filter_params)

        self.mock_query_executor_instance.get_audit_log_entries.assert_called_once_with(**filter_params)
        self.assertEqual(retrieved_entries, expected_entries_data)

        # Test with no filters
        self.mock_query_executor_instance.reset_mock()
        expected_empty_call_entries = [AuditLogEntry(timestamp=now, app_name="app3", user_name="user3", model="model3", log_type="type3", id=3, prompt_text=None, response_text=None, remote_completion_id=None, project=None)]
        self.mock_query_executor_instance.get_audit_log_entries.return_value = expected_empty_call_entries

        retrieved_no_filters = self.backend.get_audit_log_entries()
        self.mock_query_executor_instance.get_audit_log_entries.assert_called_once_with(
            start_date=None, end_date=None, app_name=None, user_name=None,
            project=None, log_type=None, limit=None
        )
        self.assertEqual(retrieved_no_filters, expected_empty_call_entries)



class MockQueryExecutor:
    def __init__(self, parent_backend):
        self.parent_backend = parent_backend

    def execute_query(self, query: str) -> list[dict]:
        """Mocks executing a raw SQL SELECT query."""
        print(f"MockBackend: Executing query: {query}")
        if query.strip().upper().startswith("SELECT"):
            return [
                {"id": 1, "model": "mock_model_A", "tokens": 100},
                {"id": 2, "model": "mock_model_B", "tokens": 200},
            ]
        raise ValueError("MockBackend only supports SELECT queries for execute_query.")



import unittest
from unittest.mock import patch, MagicMock, call
import os
from datetime import datetime, timezone, timedelta

from src.llm_accounting.backends.postgresql import PostgreSQLBackend
from src.llm_accounting.backends.base import UsageEntry, UsageStats, AuditLogEntry
from src.llm_accounting.models.limits import UsageLimitDTO, LimitScope, LimitType, TimeInterval
from typing import List, Optional

import psycopg2

from tests.backends.postgresql_backend_tests.base_test_postgresql import BaseTestPostgreSQL


class TestPostgreSQLQueryExecution(BaseTestPostgreSQL):

    def test_execute_query_success(self):
        self.backend.initialize()
        raw_query_for_execute = "SELECT * FROM accounting_entries WHERE cost > 10;"
        expected_result = [{'id': 1, 'cost': 20.0}, {'id': 2, 'cost': 30.0}]
        self.mock_cursor.fetchall.return_value = expected_result
        
        result = self.backend.execute_query(raw_query_for_execute)

        self.mock_cursor.execute.assert_called_once_with(raw_query_for_execute)
        self.assertEqual(result, expected_result)

    def test_execute_query_non_select_error(self):
        self.backend.initialize()
        non_select_query = "DELETE FROM accounting_entries;"
        with self.assertRaisesRegex(ValueError, "Only SELECT queries are allowed"):
            self.backend.execute_query(non_select_query)
        self.mock_cursor.execute.assert_not_called()



import sqlite3 # Keep for type hints if any raw sqlite3 usage remains by mistake, but should be phased out
import typing
from unittest.mock import MagicMock, patch

import pytest
from sqlalchemy import text # Added import

from llm_accounting.backends.sqlite import SQLiteBackend


@pytest.fixture
def test_db():
    """Fixture setting up an in-memory test database with sample data using SQLAlchemy connection."""
    # Use a unique in-memory database for each test to avoid locking issues
    backend = SQLiteBackend(":memory:") # Uses "sqlite:///:memory:" for SQLAlchemy engine
    backend.initialize() # This initializes self.engine and self.conn (SQLAlchemy connection)
    
    # conn is now an SQLAlchemy Connection object
    conn = backend.conn 
    assert conn is not None, "SQLAlchemy connection not initialized in SQLiteBackend"

    create_table_sql_script = """
        CREATE TABLE IF NOT EXISTS accounting_entries (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            timestamp TEXT NOT NULL DEFAULT CURRENT_TIMESTAMP,
            model TEXT NOT NULL,
            prompt_tokens INTEGER,
            completion_tokens INTEGER,
            total_tokens INTEGER,
            local_prompt_tokens INTEGER,
            local_completion_tokens INTEGER,
            local_total_tokens INTEGER,
            cost REAL NOT NULL,
            execution_time REAL NOT NULL,
            caller_name TEXT NOT NULL DEFAULT '',
            username TEXT NOT NULL DEFAULT '',
            cached_tokens INTEGER NOT NULL DEFAULT 0,
            reasoning_tokens INTEGER NOT NULL DEFAULT 0
        );
    """
    # Execute DDL (CREATE TABLE)
    # DDL statements usually don't need to be part of an explicit transaction commit with SQLAlchemy,
    # or are auto-committed by some drivers/DBs.
    conn.execute(text(create_table_sql_script))

    insert_sql_template = text("""
        INSERT INTO accounting_entries 
        (model, username, timestamp, prompt_tokens, completion_tokens, total_tokens, cost, execution_time, cached_tokens, reasoning_tokens) 
        VALUES (:model, :username, :timestamp, :prompt_tokens, :completion_tokens, :total_tokens, :cost, :execution_time, :cached_tokens, :reasoning_tokens)
    """)
    
    data_to_insert = [
        {"model": "gpt-4", "username": "user1", "timestamp": "2024-01-01 10:00", "prompt_tokens": 100, "completion_tokens": 150, "total_tokens": 250, "cost": 0.06, "execution_time": 1.5, "cached_tokens": 0, "reasoning_tokens": 0},
        {"model": "gpt-4", "username": "user2", "timestamp": "2024-01-01 11:00", "prompt_tokens": 150, "completion_tokens": 100, "total_tokens": 250, "cost": 0.09, "execution_time": 2.1, "cached_tokens": 0, "reasoning_tokens": 0},
        {"model": "gpt-3.5", "username": "user1", "timestamp": "2024-01-01 12:00", "prompt_tokens": 50, "completion_tokens": 75, "total_tokens": 125, "cost": 0.002, "execution_time": 0.8, "cached_tokens": 0, "reasoning_tokens": 0},
        {"model": "gpt-3.5", "username": "user3", "timestamp": "2024-01-01 13:00", "prompt_tokens": 75, "completion_tokens": 50, "total_tokens": 125, "cost": 0.003, "execution_time": 1.2, "cached_tokens": 0, "reasoning_tokens": 0},
    ]
    
    # Execute DML (INSERT)
    conn.execute(insert_sql_template, data_to_insert)
    conn.commit() # Ensure data is committed to be visible for tests

    count_result = conn.execute(text('SELECT COUNT(*) FROM accounting_entries')).scalar_one()
    print(f"Data inserted. Rows in accounting_entries: {count_result}")
    return backend


@pytest.fixture(autouse=True)
def mock_get_accounting(test_db):
    """
    Fixture to patch llm_accounting.cli.get_accounting to return a mock LLMAccounting
    instance that uses our test_db backend. This ensures that CLI commands use the
    in-memory database and that mock calls are properly tracked.
    """
    with patch('llm_accounting.cli.utils.get_accounting') as mock_get_acc:
        mock_accounting_instance = MagicMock()
        mock_accounting_instance.backend = test_db
        mock_accounting_instance.__enter__.return_value = mock_accounting_instance
        mock_accounting_instance.__exit__.return_value = None  # Ensure __exit__ is callable and returns None

        # Explicitly mock methods that are called on the LLMAccounting instance
        # and delegate them to the test_db backend
        mock_accounting_instance.get_period_stats.side_effect = test_db.get_period_stats
        mock_accounting_instance.get_model_stats.side_effect = test_db.get_model_stats
        mock_accounting_instance.get_model_rankings.side_effect = test_db.get_model_rankings
        mock_accounting_instance.purge.side_effect = test_db.purge
        mock_accounting_instance.tail.side_effect = test_db.tail
        mock_accounting_instance.track_usage.side_effect = test_db.insert_usage  # LLMAccounting.track_usage calls backend.insert_usage
        # For CLI select, the backend's execute_query is called by a helper in cli.commands.select_cmd
        # The line below was causing an AttributeError and is removed.
        # mock_accounting_instance.backend.execute_query.side_effect = test_db.execute_query
        # Assigning mock_accounting_instance.backend = test_db is sufficient for calls to
        # mock_accounting_instance.backend.execute_query() to use the method from the test_db instance.


        mock_get_acc.return_value = mock_accounting_instance
        yield mock_get_acc



from datetime import datetime
from typing import List, Optional
# from typing_extensions import override # Removed as it's not directly overriding BaseBackend
from llm_accounting.models.limits import LimitScope, LimitType, UsageLimitDTO  # Corrected import path


class MockLimitManager:
    def __init__(self, parent_backend):
        self.parent_backend = parent_backend

    def insert_usage_limit(self, limit: UsageLimitDTO) -> None:
        """Mocks inserting a usage limit."""
        if limit.id is None:
            limit.id = self.parent_backend.next_limit_id
            self.parent_backend.next_limit_id += 1
        self.parent_backend.limits.append(limit)
        print(f"MockBackend: Inserted usage limit for scope {limit.scope} with ID {limit.id}")

    def delete_usage_limit(self, limit_id: int) -> None:
        """Mocks deleting a usage limit."""
        initial_len = len(self.parent_backend.limits)
        self.parent_backend.limits = [limit for limit in self.parent_backend.limits if limit.id != limit_id]
        if len(self.parent_backend.limits) < initial_len:
            print(f"MockBackend: Deleted usage limit with ID {limit_id}")
        else:
            print(f"MockBackend: No usage limit found with ID {limit_id} to delete.")

    def get_usage_limits(
        self,
        scope: Optional[LimitScope] = None,
        model: Optional[str] = None,
        username: Optional[str] = None,
        caller_name: Optional[str] = None,
        project_name: Optional[str] = None,
        filter_project_null: Optional[bool] = False,
        filter_username_null: Optional[bool] = False,
        filter_caller_name_null: Optional[bool] = False,
    ) -> List[UsageLimitDTO]:
        """Mocks retrieving usage limits."""
        print(f"MockBackend: Getting usage limits with filters: scope={scope}, model={model}, username={username}, caller_name={caller_name}, project_name={project_name}, filter_project_null={filter_project_null}, filter_username_null={filter_username_null}, filter_caller_name_null={filter_caller_name_null}")

        filtered_limits = self.parent_backend.limits

        if scope:
            filtered_limits = [limit for limit in filtered_limits if limit.scope == scope.value]
        if model:
            filtered_limits = [limit for limit in filtered_limits if limit.model == model]
        if username:
            filtered_limits = [limit for limit in filtered_limits if limit.username == username]
        if caller_name:
            filtered_limits = [limit for limit in filtered_limits if limit.caller_name == caller_name]
        if project_name:
            filtered_limits = [limit for limit in filtered_limits if limit.project_name == project_name]

        if filter_project_null:
            filtered_limits = [limit for limit in filtered_limits if limit.project_name is None]
        if filter_username_null:
            filtered_limits = [limit for limit in filtered_limits if limit.username is None]
        if filter_caller_name_null:
            filtered_limits = [limit for limit in filtered_limits if limit.caller_name is None]

        return filtered_limits

    def get_accounting_entries_for_quota(
        self,
        start_time: datetime,
        limit_type: LimitType,
        model: Optional[str] = None,
        username: Optional[str] = None,
        caller_name: Optional[str] = None,
        project_name: Optional[str] = None,
        filter_project_null: Optional[bool] = False,
    ) -> float:
        """
        Mocks getting accounting entries for quota calculation.
        """
        print(f"MockBackend: Getting accounting entries for quota (type: {limit_type.value}) from {start_time} with filters: model={model}, username={username}, caller_name={caller_name}, project_name={project_name}, filter_project_null={filter_project_null}")
        mock_value = 100.0
        if limit_type == LimitType.REQUESTS:
            mock_value = 10.0
        elif limit_type == LimitType.COST:
            mock_value = 5.0

        if model == "specific_model_for_quota_test":
            mock_value /= 2

        return mock_value



import sys
from io import StringIO
from unittest.mock import MagicMock, patch

import pytest

from llm_accounting.cli.main import main as cli_main


@patch("llm_accounting.cli.utils.get_accounting")
def test_select_no_results(mock_get_accounting, test_db, capsys):
    """Test query that returns no results"""
    mock_accounting_instance = MagicMock()
    mock_get_accounting.return_value = mock_accounting_instance
    mock_accounting_instance.__enter__.return_value = mock_accounting_instance
    mock_accounting_instance.__exit__.return_value = None
    mock_accounting_instance.backend.execute_query.return_value = []

    with patch.object(sys, 'argv', ['cli_main', "select", "--query", "SELECT * FROM accounting_entries WHERE username = 'nonexistent'"]):
        cli_main()

    captured = capsys.readouterr()
    assert "No results found" in captured.out
    mock_accounting_instance.__exit__.assert_called_once()



import io
import contextlib
import pytest
from importlib.metadata import version
from llm_accounting.cli.main import main

def test_cli_version(monkeypatch):
    expected_version = version('llm-accounting')
    
    # Monkeypatch sys.argv and capture exit/output
    with monkeypatch.context() as m:
        m.setattr('sys.argv', ['llm-accounting', '--version'])
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            with pytest.raises(SystemExit) as exc_info:
                main()
            
    assert exc_info.value.code == 0
    output = f.getvalue().strip()
    assert output == f'llm-accounting {expected_version}'



from datetime import datetime
from typing import List

from ..base import UsageEntry


class MockUsageManager:
    def __init__(self, parent_backend):
        self.parent_backend = parent_backend

    def insert_usage(self, entry: UsageEntry) -> None:
        """Mocks inserting a new usage entry."""
        self.parent_backend.entries.append(entry)
        print(f"MockBackend: Inserted usage for model {entry.model}")  # noqa: T201

    def purge(self) -> None:
        """Mocks deleting all usage entries."""
        self.parent_backend.entries = []
        self.parent_backend.limits = []
        print("MockBackend: All usage entries and limits purged.")  # noqa: T201

    def tail(self, n: int = 10) -> List[UsageEntry]:
        """Mocks getting the n most recent usage entries."""
        print(f"MockBackend: Getting last {n} usage entries.")  # noqa: T201
        if not self.parent_backend.entries:
            return [
                UsageEntry(
                    model="mock_model_1",
                    prompt_tokens=10,
                    completion_tokens=20,
                    total_tokens=30,
                    cost=0.01,
                    execution_time=0.05,
                    timestamp=datetime.now()
                ),
                UsageEntry(
                    model="mock_model_2",
                    prompt_tokens=15,
                    completion_tokens=25,
                    total_tokens=40,
                    cost=0.02,
                    execution_time=0.08,
                    timestamp=datetime.now()
                ),
            ][:n]
        return self.parent_backend.entries[-n:]



from datetime import datetime, timezone

import pytest

from llm_accounting import LLMAccounting
from llm_accounting.backends.sqlite import SQLiteBackend
from llm_accounting.models.limits import (LimitScope, LimitType, TimeInterval,
                                          UsageLimitDTO)


@pytest.fixture
def sqlite_backend_for_accounting(temp_db_path):
    """Create and initialize a SQLite backend for LLMAccounting"""
    backend = SQLiteBackend(db_path=temp_db_path)
    backend.initialize()
    yield backend
    backend.close()


@pytest.fixture
def accounting_instance(sqlite_backend_for_accounting):
    """Create an LLMAccounting instance with a temporary SQLite backend"""
    acc = LLMAccounting(backend=sqlite_backend_for_accounting)
    acc.__enter__() 
    yield acc
    acc.__exit__(None, None, None)


def test_global_limit(accounting_instance: LLMAccounting, sqlite_backend_for_accounting: SQLiteBackend):
    # Use the backend directly to add UsageLimitData for setup
    limit_to_set = UsageLimitDTO(
        scope=LimitScope.GLOBAL.value,
        limit_type=LimitType.REQUESTS.value,
        max_value=10,
        interval_unit=TimeInterval.MINUTE.value,
        interval_value=1
    )
    sqlite_backend_for_accounting.insert_usage_limit(limit_to_set)

    # Check and add requests sequentially using accounting_instance
    for i in range(10):
        allowed, reason = accounting_instance.check_quota("gpt-4", "user1", "app1", 1000, 0.25)
        assert allowed, f"Request {i+1}/10 should be allowed. Reason: {reason}"
        accounting_instance.track_usage(
            model="gpt-4",
            username="user1",
            caller_name="app1",
            prompt_tokens=1000,
            completion_tokens=500,
            cost=0.25,
            timestamp=datetime.now(timezone.utc)
        )

    # Add 11th request to exceed limit
    allowed, message = accounting_instance.check_quota("gpt-4", "user1", "app1", 1000, 0.25)
    assert not allowed, "11th request should be denied"
    assert message is not None, "Denial message should not be None"
    
    expected_message_part_1 = "GLOBAL limit: 10.00 requests per 1 minute"
    expected_message_part_2 = "current usage: 10.00, request: 1.00"
    
    assert expected_message_part_1 in message
    assert expected_message_part_2 in message



import argparse
import sys
import os
import platform
from importlib.metadata import version as get_version

from .parsers import (add_purge_parser, add_select_parser, add_stats_parser,
                      add_tail_parser, add_track_parser, add_limits_parser)
from .utils import console


def _check_privileged_user():
    """
    Checks if the current user is a privileged user (root on Linux/macOS, admin on Windows).
    Exits the program with an error message if the user is privileged.
    """
    if platform.system() == "Windows":
        try:
            # Check if user has admin privileges on Windows
            import ctypes
            if ctypes.windll.shell32.IsUserAnAdmin():
                console.print("[red]Error: Running the CLI as an administrator is not allowed for security reasons.[/red]")
                sys.exit(1)
        except AttributeError:
            # Handle cases where ctypes might not be available or IsUserAnAdmin fails
            pass
    elif hasattr(os, 'geteuid') and os.geteuid() == 0:  # type: ignore
        # Check for root user on Linux/macOS
        console.print("[red]Error: Running the CLI as root is not allowed for security reasons.[/red]")
        sys.exit(1)

def main():
    _check_privileged_user()
    package_version = get_version('llm-accounting')
    parser = argparse.ArgumentParser(
        description="LLM Accounting CLI - Track and analyze LLM usage",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument('--version', action='version', version=f'llm-accounting {package_version}')
    parser.add_argument(
        "--db-file",
        type=str,
        help="SQLite database file path (must end with .sqlite, .sqlite3 or .db). "
             "Only applicable when --db-backend is 'sqlite'.",
    )
    parser.add_argument(
        "--db-backend",
        type=str,
        default="sqlite",
        choices=["sqlite", "postgresql"],
        help="Select the database backend (sqlite or postgresql). Defaults to 'sqlite'.",
    )
    parser.add_argument(
        "--postgresql-connection-string",
        type=str,
        help="Connection string for the PostgreSQL database. "
             "Required when --db-backend is 'postgresql'. "
             "Can also be provided via POSTGRESQL_CONNECTION_STRING environment variable.",
    )
    parser.add_argument(
        "--project-name",
        type=str,
        help="Default project name to associate with usage entries. Can be overridden by command-specific --project.",
    )
    parser.add_argument(
        "--app-name",
        type=str,
        help="Default application name to associate with usage entries. Can be overridden by command-specific --caller-name.",
    )
    parser.add_argument(
        "--user-name",
        type=str,
        help="Default user name to associate with usage entries. Can be overridden by command-specific --username. Defaults to current system user.",
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    add_stats_parser(subparsers)
    add_purge_parser(subparsers)
    add_tail_parser(subparsers)
    add_select_parser(subparsers)
    add_track_parser(subparsers)
    add_limits_parser(subparsers)

    args = parser.parse_args()

    if args.command is None:
        parser.print_help()
        sys.exit(0)

    from .utils import get_accounting

    try:
        accounting = get_accounting(
            db_backend=args.db_backend,
            db_file=args.db_file,
            postgresql_connection_string=args.postgresql_connection_string,
            project_name=args.project_name,
            app_name=args.app_name,
            user_name=args.user_name,
        )
        with accounting:
            args.func(args, accounting)
    except SystemExit:
        # Allow SystemExit to propagate, especially for pytest.raises
        raise
    except Exception as e:
        console.print(f"[red]Error: {e}[/red]")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Main package initialization for LLM Accounting system.

This package provides core functionality for tracking and managing API usage quotas
and rate limits across multiple services.
"""
import logging
import os # For run_migrations
from datetime import datetime
from typing import Dict, List, Optional, Tuple

from alembic.config import Config as AlembicConfig # For run_migrations
from alembic import command as alembic_command # For run_migrations

from .backends.base import BaseBackend, UsageEntry, UsageStats
from .backends.mock_backend import MockBackend
from .backends.sqlite import SQLiteBackend
from .models.limits import LimitScope, LimitType, TimeInterval, UsageLimitDTO
from .services.quota_service import QuotaService
from .audit_log import AuditLogger

logger = logging.getLogger(__name__)

# Alembic migration function
def run_migrations():
    """Checks and applies any pending database migrations."""
    # Use the same logger as the rest of the llm_accounting package
    migration_logger = logging.getLogger(__name__ + ".migrations") 
    
    db_url = os.getenv("LLM_ACCOUNTING_DB_URL")
    alembic_ini_path = "alembic.ini" # Assuming it's in the project root

    if not db_url:
        # If LLM_ACCOUNTING_DB_URL is not set, try to get from alembic.ini as a fallback
        migration_logger.info(f"LLM_ACCOUNTING_DB_URL not set. Attempting to read from {alembic_ini_path}.")
        try:
            # Create a temporary config to read the URL, don't use this for upgrade command
            temp_config = AlembicConfig(alembic_ini_path)
            db_url = temp_config.get_main_option("sqlalchemy.url", None)
            if db_url:
                migration_logger.info(f"Found sqlalchemy.url in {alembic_ini_path}: {db_url[:db_url.find('://')+3]}...")
            else:
                migration_logger.warning(f"sqlalchemy.url not found in {alembic_ini_path}.")
        except Exception as e:
            migration_logger.warning(f"Could not read sqlalchemy.url from {alembic_ini_path} for migrations: {e}")

    if not db_url:
        migration_logger.error("Database URL for migrations not found. Skipping migrations. Set LLM_ACCOUNTING_DB_URL or ensure sqlalchemy.url is in alembic.ini.")
        return

    # Log minimally, avoiding printing username/password if present in URL
    log_db_url = db_url
    try:
        from sqlalchemy.engine.url import make_url
        parsed_url = make_url(db_url)
        if parsed_url.password:
            log_db_url = str(parsed_url._replace(password="****"))
    except Exception:
        # If parsing fails, log the original URL but be mindful
        pass 
    migration_logger.info(f"Running database migrations for URL: {log_db_url}")
    
    try:
        alembic_cfg = AlembicConfig(alembic_ini_path)
        # Crucially, override the sqlalchemy.url from alembic.ini with the one we've determined
        alembic_cfg.set_main_option("sqlalchemy.url", db_url)
        
        alembic_command.upgrade(alembic_cfg, "head")
        migration_logger.info("Database migrations are up to date.")
    except Exception as e:
        migration_logger.error(f"Error running database migrations: {e}", exc_info=True)
        # Depending on the application's needs, you might want to re-raise or exit
        # For now, we log and continue, allowing the app to attempt to start.
        # Consider making this critical based on application requirements.
        # raise


class LLMAccounting:
    """Main interface for LLM usage tracking"""

    def __init__(
        self,
        backend: Optional[BaseBackend] = None,
        project_name: Optional[str] = None,
        app_name: Optional[str] = None,
        user_name: Optional[str] = None,
    ):
        """Initialize with an optional backend. If none provided, uses SQLiteBackend."""
        
        # Run database migrations before backend initialization
        run_migrations()
        
        self.backend = backend or SQLiteBackend()
        self.quota_service = QuotaService(self.backend)
        self.project_name = project_name
        self.app_name = app_name
        self.user_name = user_name

    def __enter__(self):
        """Initialize the backend when entering context"""
        logger.info("Entering LLMAccounting context.")
        self.backend.initialize()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Close the backend when exiting context"""
        logger.info("Exiting LLMAccounting context. Closing backend.")
        self.backend.close()
        if exc_type:
            logger.error(
                f"LLMAccounting context exited with exception: {exc_type.__name__}: {exc_val}"
            )

    def track_usage(
        self,
        model: str,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        local_prompt_tokens: Optional[int] = None,
        local_completion_tokens: Optional[int] = None,
        local_total_tokens: Optional[int] = None,
        cost: float = 0.0,
        execution_time: float = 0.0,
        timestamp: Optional[datetime] = None,
        caller_name: Optional[str] = None, # Changed to Optional[str]
        username: Optional[str] = None, # Changed to Optional[str]
        cached_tokens: int = 0,
        reasoning_tokens: int = 0,
        project: Optional[str] = None,
    ) -> None:
        """Track a new LLM usage entry"""
        self.backend._ensure_connected()
        entry = UsageEntry(
            model=model,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=total_tokens,
            local_prompt_tokens=local_prompt_tokens,
            local_completion_tokens=local_completion_tokens,
            local_total_tokens=local_total_tokens,
            cost=cost,
            execution_time=execution_time,
            timestamp=timestamp,
            caller_name=caller_name if caller_name is not None else self.app_name, # Use instance default
            username=username if username is not None else self.user_name, # Use instance default
            cached_tokens=cached_tokens,
            reasoning_tokens=reasoning_tokens,
            project=project if project is not None else self.project_name, # Use instance default
        )
        self.backend.insert_usage(entry)

    def get_period_stats(self, start: datetime, end: datetime) -> UsageStats:
        """Get aggregated statistics for a time period"""
        self.backend._ensure_connected()
        return self.backend.get_period_stats(start, end)

    def get_model_stats(self, start: datetime, end: datetime):
        """Get statistics grouped by model for a time period"""
        self.backend._ensure_connected()
        return self.backend.get_model_stats(start, end)

    def get_model_rankings(
        self, start_date: datetime, end_date: datetime
    ) -> Dict[str, List[Tuple[str, float]]]:
        """Get model rankings based on different metrics"""
        self.backend._ensure_connected()
        return self.backend.get_model_rankings(start_date, end_date)

    def purge(self) -> None:
        """Delete all usage entries from the backend"""
        self.backend._ensure_connected()
        self.backend.purge()

    def tail(self, n: int = 10) -> List[UsageEntry]:
        """Get the n most recent usage entries"""
        self.backend._ensure_connected()
        return self.backend.tail(n)

    def check_quota(
        self,
        model: str,
        username: Optional[str],
        caller_name: Optional[str],
        input_tokens: int,
        cost: float = 0.0,
        project_name: Optional[str] = None,
    ) -> Tuple[bool, Optional[str]]:
        """Check if the current request exceeds any defined quotas."""
        self.backend._ensure_connected()
        return self.quota_service.check_quota(
            model=model,
            username=username,
            caller_name=caller_name,
            input_tokens=input_tokens,
            cost=cost,
            project_name=project_name
        )

    def set_usage_limit(
        self,
        scope: LimitScope,
        limit_type: LimitType,
        max_value: float,
        interval_unit: TimeInterval,
        interval_value: int,
        model: Optional[str] = None,
        username: Optional[str] = None,
        caller_name: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        """Sets a new usage limit."""
        self.backend._ensure_connected()
        limit = UsageLimitDTO(
            scope=scope.value if isinstance(scope, LimitScope) else scope,
            limit_type=limit_type.value if isinstance(limit_type, LimitType) else limit_type,
            max_value=max_value,
            interval_unit=interval_unit.value if isinstance(interval_unit, TimeInterval) else interval_unit,
            interval_value=interval_value,
            model=model,
            username=username,
            caller_name=caller_name,
            project_name=project_name,
        )
        self.backend.insert_usage_limit(limit)

    def get_usage_limits(
        self,
        scope: Optional[LimitScope] = None,
        model: Optional[str] = None,
        username: Optional[str] = None,
        caller_name: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> List[UsageLimitDTO]:
        """Retrieves configured usage limits."""
        self.backend._ensure_connected()
        return self.backend.get_usage_limits(
            scope=scope,
            model=model,
            username=username,
            caller_name=caller_name,
            project_name=project_name
        )

    def delete_usage_limit(self, limit_id: int) -> None:
        """Deletes a usage limit by its ID."""
        self.backend._ensure_connected()
        self.backend.delete_usage_limit(limit_id)

    def get_db_path(self) -> Optional[str]:
        """
        Returns the database path if the backend is a SQLiteBackend.
        Otherwise, returns None.
        """
        if isinstance(self.backend, SQLiteBackend):
            return self.backend.db_path
        return None


# Export commonly used classes
__all__ = [
    "LLMAccounting",
    "BaseBackend",
    "UsageEntry",
    "UsageStats",
    "SQLiteBackend",
    "MockBackend",
    "AuditLogger",
    "LimitScope",
    "LimitType",
    "TimeInterval",
    "UsageLimitDTO",
]



from .sqlite import SQLiteBackend


def get_backend() -> SQLiteBackend:
    """Get the configured backend instance"""
    return SQLiteBackend("file:memdb.sqlite?mode=memory&cache=shared")


__all__ = ["SQLiteBackend"]



import subprocess
import os
import sys
import time
from datetime import datetime

# Define the base directory for tests
TESTS_DIR = "tests"
# Timeout for each test in seconds
TEST_TIMEOUT = 15


def get_test_functions():
    """
    Dynamically discovers test functions from specified directories.
    This is a placeholder for the actual discovery logic.
    In a real scenario, this would parse the output of `pytest --collect-only`.
    For now, we'll hardcode based on the previous `list_code_definition_names` output.
    """
    test_functions = []

    # From tests/backends/
    test_functions.extend([
        "tests/backends/test_base.py::test_backend_interface",
        "tests/backends/test_base.py::test_backend_abstract_methods",
        "tests/backends/test_sqlite.py::test_initialize",
        "tests/backends/test_sqlite.py::test_insert_usage",
        "tests/backends/test_sqlite.py::test_get_period_stats",
        "tests/backends/test_sqlite.py::test_get_model_stats",
        "tests/backends/test_sqlite.py::test_get_model_rankings",
        "tests/backends/test_sqlite.py::test_purge",
        "tests/backends/test_sqlite.py::test_purge_empty_database",
        "tests/backends/test_usage_models.py::test_usage_entry_creation",
        "tests/backends/test_usage_models.py::test_usage_stats_creation",
    ])

    # From tests/cli/
    test_functions.extend([
        "tests/cli/test_cli_purge.py::test_purge_with_confirmation",
        "tests/cli/test_cli_purge.py::test_purge_without_confirmation",
        "tests/cli/test_cli_purge.py::test_purge_with_yes_flag",
        "tests/cli/test_cli_purge.py::test_purge_with_yes_flag_long",
        "tests/cli/test_cli_stats.py::test_stats_no_period",
        "tests/cli/test_cli_stats.py::test_stats_periods",
        "tests/cli/test_cli_stats.py::test_stats_custom_period",
        "tests/cli/test_cli_stats.py::test_custom_db_file_usage",
        "tests/cli/test_cli_stats.py::test_default_db_file_usage",
        "tests/cli/test_cli_stats.py::test_db_file_validation_error",
        "tests/cli/test_cli_stats.py::test_db_file_permission_error",
        "tests/cli/test_cli_tail.py::test_tail_default",
        "tests/cli/test_cli_tail.py::test_tail_custom_number",
        "tests/cli/test_cli_tail.py::test_tail_empty",
        "tests/cli/test_cli_track.py::test_track_usage_with_protected_db_file",
        "tests/cli/test_cli_track.py::test_track_usage",
        "tests/cli/test_cli_track.py::test_track_usage_with_timestamp",
        "tests/cli/test_cli_track.py::test_track_usage_with_caller_name",
        "tests/cli/test_cli_track.py::test_track_usage_with_username",
        "tests/cli/test_cli_track.py::test_track_usage_with_cached_tokens",
        "tests/cli/test_cli_track.py::test_track_usage_with_reasoning_tokens",
    ])

    # From tests/core/
    test_functions.extend([
        "tests/core/test_accounting_purge.py::test_purge",
        "tests/core/test_accounting_stats.py::test_get_period_stats",
        "tests/core/test_accounting_stats.py::test_get_model_stats",
        "tests/core/test_accounting_stats.py::test_get_model_rankings",
        "tests/core/test_accounting_tracking.py::test_track_usage",
        "tests/core/test_accounting_tracking.py::test_tail",
        "tests/core/test_accounting_tracking.py::test_tail_empty",
        "tests/core/test_accounting_tracking.py::test_tail_default_limit",
        "tests/core/test_accounting_tracking.py::test_track_usage_with_caller_and_user",
        "tests/core/test_accounting_tracking.py::test_tail_with_caller_and_user",
        "tests/core/test_accounting_validation.py::test_track_usage_empty_model",
        "tests/core/test_accounting_validation.py::test_track_usage_none_model",
        "tests/core/test_accounting_validation.py::test_usage_entry_empty_model",
        "tests/core/test_accounting_validation.py::test_usage_entry_none_model",
        "tests/core/test_accounting_validation.py::test_track_usage_without_timestamp",
        "tests/core/test_accounting_validation.py::test_track_usage_with_timestamp",
        "tests/core/test_accounting_validation.py::test_track_usage_with_token_details",
        "tests/core/test_accounting_validation.py::test_track_usage_default_token_details",
    ])

    return test_functions


def run_single_test(test_path):
    """Runs a single pytest and returns its status."""
    print(f"[{datetime.now().strftime('%H:%M:%S')}] Running: {test_path}")
    sys.stdout.flush()

    # Determine the correct python executable for the venv
    if sys.platform == "win32":
        python_executable = os.path.join(".venv", "Scripts", "python.exe")
    else:
        python_executable = os.path.join(".venv", "bin", "python")

    if not os.path.exists(python_executable):
        print(f"Error: Virtual environment not found at {python_executable}. Please activate or create it.")
        return "ERROR_VENV_NOT_FOUND"

    command = [
        python_executable,
        "-m",
        "pytest",
        "--maxfail=1",  # As per .clinerules
        "--verbose",
        "--cov=llm_accounting",
        "--cov-report=term-missing",
        "--cov-report=html",
        "--no-cov-on-fail",
        test_path
    ]

    try:
        start_time = time.time()
        process = subprocess.run(
            command,
            capture_output=True,
            text=True,
            timeout=TEST_TIMEOUT,
            check=False  # Do not raise CalledProcessError for non-zero exit codes
        )
        end_time = time.time()
        duration = end_time - start_time

        if process.returncode != 0:
            if "failed" in process.stdout.lower() or "error" in process.stderr.lower():
                print(f"  -> FAILED (Exit Code: {process.returncode}, Duration: {duration:.2f}s)")
                sys.stdout.flush()
                print("--- STDOUT ---")
                sys.stdout.flush()
                print(process.stdout)
                sys.stdout.flush()
                print("--- STDERR ---")
                sys.stdout.flush()
                print(process.stderr)
                sys.stdout.flush()
                return "FAILED"
            else:
                print(f"  -> COMPLETED (Exit Code: {process.returncode}, Duration: {duration:.2f}s)")
                sys.stdout.flush()
                print("--- STDOUT ---")
                sys.stdout.flush()
                print(process.stdout)
                sys.stdout.flush()
                print("--- STDERR ---")
                sys.stdout.flush()
                print(process.stderr)
                sys.stdout.flush()
                return "COMPLETED"
        else:
            print(f"  -> PASSED (Duration: {duration:.2f}s)")
            sys.stdout.flush()
            return "PASSED"

    except subprocess.TimeoutExpired:
        print(f"  -> HUNG (Timeout after {TEST_TIMEOUT}s)")
        sys.stdout.flush()
        return "HUNG"
    except Exception as e:
        print(f"  -> ERROR: {e}")
        sys.stdout.flush()
        return "ERROR"


def main():
    all_tests = get_test_functions()
    total_tests = len(all_tests)
    print(f"Discovered {total_tests} individual tests.")
    sys.stdout.flush()
    print("-" * 50)
    sys.stdout.flush()

    results = {
        "PASSED": [],
        "FAILED": [],
        "HUNG": [],
        "ERROR": [],
        "COMPLETED": [],  # For tests that don't pass but also don't explicitly fail or hang
        "ERROR_VENV_NOT_FOUND": []
    }

    for i, test in enumerate(all_tests):
        print(f"Processing test {i + 1}/{total_tests}: {test}")
        sys.stdout.flush()
        status = run_single_test(test)
        results[status].append(test)
        print("-" * 50)
        sys.stdout.flush()

    print("\n--- Summary ---")
    sys.stdout.flush()
    print(f"Total Tests: {total_tests}")
    sys.stdout.flush()
    print(f"Passed: {len(results['PASSED'])}")
    sys.stdout.flush()
    print(f"Failed: {len(results['FAILED'])}")
    sys.stdout.flush()
    print(f"Hung (Timeout): {len(results['HUNG'])}")
    sys.stdout.flush()
    print(f"Completed (Non-zero exit, not failed): {len(results['COMPLETED'])}")
    sys.stdout.flush()
    print(f"Errors: {len(results['ERROR'])}")
    sys.stdout.flush()
    print(f"Venv Not Found Errors: {len(results['ERROR_VENV_NOT_FOUND'])}")
    sys.stdout.flush()

    if results['HUNG']:
        print("\n--- Tests that HUNG ---")
        sys.stdout.flush()
        for test in results['HUNG']:
            print(f"- {test}")
            sys.stdout.flush()

    if results['FAILED']:
        print("\n--- Tests that FAILED ---")
        sys.stdout.flush()
        for test in results['FAILED']:
            print(f"- {test}")
            sys.stdout.flush()

    if results['ERROR'] or results['ERROR_VENV_NOT_FOUND']:
        print("\n--- Tests with ERRORS ---")
        sys.stdout.flush()
        for test in results['ERROR']:
            print(f"- {test}")
            sys.stdout.flush()
        for test in results['ERROR_VENV_NOT_FOUND']:
            print(f"- {test} (Venv Error)")
            sys.stdout.flush()


if __name__ == "__main__":
    main()
